from concurrent.futures import ThreadPoolExecutor
from asyncio import get_event_loop, gather


def call_parallel(functions):
    """
    Call functions in multiple threads.

    Create a pool of thread as large as the number of functions.
    Functions should accept no parameters (wrap then with partial or lambda).
    """
    loop = get_event_loop()
    executor = ThreadPoolExecutor(max_workers=len(functions))

    try:
        tasks = [
            loop.run_in_executor(executor, function)
            for function in functions
        ]
        result = loop.run_until_complete(gather(*tasks))

    finally:
        loop.close()

    return result



import json
import struct
import sys
from abc import ABC
from abc import abstractmethod
from typing import BinaryIO
from typing import Union

from brotab.inout import TimeoutIO
from brotab.mediator.log import mediator_logger


class Transport(ABC):
    @abstractmethod
    def send(self, command: dict) -> None:
        pass

    @abstractmethod
    def recv(self) -> dict:
        pass

    @abstractmethod
    def close(self) -> None:
        pass


def default_transport() -> Transport:
    return StdTransport(sys.stdin.buffer, sys.stdout.buffer)


def transport_with_timeout(input_, output: Union[BinaryIO, int], timeout: float) -> Transport:
    return StdTransport(TimeoutIO(input_, timeout),
                        TimeoutIO(output, timeout))


class TransportError(Exception):
    pass


class StdTransport(Transport):
    def __init__(self, input_file: BinaryIO, output_file: BinaryIO):
        self._in = input_file
        self._out = output_file

    def reset(self):
        self._in.seek(0)
        self._out.seek(0)

    def send(self, command: dict) -> None:
        encoded = self._encode(command)
        mediator_logger.info('StdTransport SENDING: %s', command)
        self._out.write(encoded['length'])
        self._out.write(encoded['content'])
        self._out.flush()
        mediator_logger.info('StdTransport SENDING DONE: %s', command)

    def recv(self) -> dict:
        mediator_logger.info('StdTransport RECEIVING')
        raw_length = self._in.read(4)
        if len(raw_length) == 0:
            raise TransportError('StdTransport: cannot read, raw_length is empty')
        message_length = struct.unpack('@I', raw_length)[0]
        message = self._in.read(message_length).decode('utf8')
        mediator_logger.info('StdTransport RECEIVED: %s', message.encode('utf8'))
        return json.loads(message)

    def _encode(self, message):
        encoded_content = json.dumps(message).encode('utf8')
        encoded_length = struct.pack('@I', len(encoded_content))
        return {'length': encoded_length, 'content': encoded_content}

    def close(self):
        self._in.close()
        self._out.close()



import os
from json import loads
from threading import Thread
from urllib.parse import unquote_plus
from wsgiref.simple_server import make_server

from flask import Flask
from flask import request

from brotab.mediator.const import DEFAULT_GET_HTML_DELIMITER_REGEX
from brotab.mediator.const import DEFAULT_GET_HTML_REPLACE_WITH
from brotab.mediator.const import DEFAULT_GET_TEXT_DELIMITER_REGEX
from brotab.mediator.const import DEFAULT_GET_TEXT_REPLACE_WITH
from brotab.mediator.const import DEFAULT_GET_WORDS_JOIN_WITH
from brotab.mediator.const import DEFAULT_GET_WORDS_MATCH_REGEX
from brotab.mediator.log import mediator_logger
from brotab.mediator.remote_api import BrowserRemoteAPI
from brotab.mediator.runner import Runner
from brotab.mediator.support import is_valid_integer
from brotab.mediator.transport import TransportError
from brotab.utils import decode_query


class MediatorHttpServer:
    def __init__(self, host: str, port: int, remote_api: BrowserRemoteAPI, poll_interval: float):
        self.host: str = host
        self.port: int = port
        self.remote_api: BrowserRemoteAPI = remote_api
        self.pid: int = os.getpid()
        self.app = Flask(__name__)
        self.http_server = make_server(host=host, port=port, app=self.app)
        self._setup_routes()

        def serve():
            mediator_logger.info('Serving mediator on %s:%s', host, port)
            self.http_server.serve_forever(poll_interval=poll_interval)

        def shutdown(join: bool):
            mediator_logger.info('Closing mediator http server on %s:%s', host, port)
            self.http_server.server_close()
            mediator_logger.info('Shutting down mediator http server on %s:%s', host, port)
            thread = Thread(target=self.http_server.shutdown)
            thread.daemon = True
            thread.start()
            if join:
                thread.join()
            mediator_logger.info('Done shutting down mediator (is_alive=%s) http server on %s:%s',
                                 thread.is_alive(), host, port)

        self.run = Runner(serve, shutdown)

    def _setup_routes(self) -> None:
        mediator_logger.info('Starting mediator http server on %s:%s pid=%s', self.host, self.port, self.pid)
        self.app.register_error_handler(ConnectionError, self.error_handler)
        self.app.register_error_handler(TimeoutError, self.error_handler)
        self.app.register_error_handler(ValueError, self.error_handler)
        self.app.register_error_handler(TransportError, self.error_handler)
        self.app.route('/', methods=['GET'])(self.root_handler)
        self.app.route('/shutdown', methods=['GET'])(self.shutdown)
        self.app.route('/list_tabs', methods=['GET'])(self.list_tabs)
        self.app.route('/query_tabs/<query_info>', methods=['GET'])(self.query_tabs)
        self.app.route('/move_tabs/<query_info>', methods=['GET'])(self.move_tabs)
        self.app.route('/open_urls/<int:window_id>', methods=['POST'])(self.open_urls)
        self.app.route('/update_tabs', methods=['POST'])(self.update_tabs)
        self.app.route('/open_urls', methods=['POST'])(self.open_urls)
        self.app.route('/close_tabs/<tab_ids>', methods=['GET'])(self.close_tabs)
        self.app.route('/new_tab/<query>', methods=['GET'])(self.new_tab)
        self.app.route('/activate_tab/<int:tab_id>', methods=['GET'])(self.activate_tab)
        self.app.route('/get_active_tabs', methods=['GET'])(self.get_active_tabs)
        self.app.route('/get_words/<string:tab_id>', methods=['GET'])(self.get_words)
        self.app.route('/get_words', methods=['GET'])(self.get_words)
        self.app.route('/get_text', methods=['GET'])(self.get_text)
        self.app.route('/get_html', methods=['GET'])(self.get_html)
        self.app.route('/get_pid', methods=['GET'])(self.get_pid)
        self.app.route('/get_browser', methods=['GET'])(self.get_browser)
        self.app.route('/echo', methods=['GET'])(self.echo)

    def error_handler(self, e: Exception):
        mediator_logger.exception('Shutting down mediator http server due to exception: %s', e)
        # can't wait for shutdown here because we're processing a request right now,
        # we will get deadlocked if we wait (join=True)
        self.run.shutdown(join=False)
        return '<ERROR>'

    def root_handler(self):
        links = []
        for rule in self.app.url_map.iter_rules():
            methods = ','.join(rule.methods)
            line = '{0}\t{1}\t{2}'.format(rule.endpoint, methods, rule)
            links.append(line)
        return '\n'.join(links)

    def shutdown(self):
        # can't wait for shutdown here because we're processing a request right now,
        # we will get deadlocked if we wait (join=True)
        self.run.shutdown(join=False)
        return 'OK'

    def list_tabs(self):
        tabs = self.remote_api.list_tabs()
        return '\n'.join(tabs)

    def query_tabs(self, query_info):
        tabs = self.remote_api.query_tabs(query_info)
        return '\n'.join(tabs)

    def move_tabs(self, move_triplets):
        return self.remote_api.move_tabs(unquote_plus(move_triplets))

    def open_urls(self, window_id=None):
        urls = request.files.get('urls')
        if urls is None:
            return 'ERROR: Please provide urls file in the request'
        urls = urls.stream.read().decode('utf8').splitlines()
        mediator_logger.info('Open urls (window_id = %s): %s', window_id, urls)
        result = self.remote_api.open_urls(urls, window_id)
        mediator_logger.info('Open urls result: %s', str(result))
        return '\n'.join(result)

    def update_tabs(self):
        updates = request.files.get('updates')
        if updates is None:
            return 'ERROR: Please provide updates in the request'
        updates = loads(updates.stream.read().decode('utf8'))
        mediator_logger.info('Sending tab updates: %s', updates)
        result = self.remote_api.update_tabs(updates)
        mediator_logger.info('Update tabs result: %s', str(result))
        return '\n'.join(result)

    def close_tabs(self, tab_ids):
        return self.remote_api.close_tabs(tab_ids)

    def new_tab(self, query):
        return self.remote_api.new_tab(query)

    def activate_tab(self, tab_id):
        focused = bool(request.args.get('focused', False))
        self.remote_api.activate_tab(tab_id, focused)
        return 'OK'

    def get_active_tabs(self):
        return self.remote_api.get_active_tabs()

    def get_words(self, tab_id=None):
        tab_id = int(tab_id) if is_valid_integer(tab_id) else None
        match_regex = request.args.get('match_regex', DEFAULT_GET_WORDS_MATCH_REGEX)
        join_with = request.args.get('join_with', DEFAULT_GET_WORDS_JOIN_WITH)
        words = self.remote_api.get_words(tab_id,
                                          decode_query(match_regex),
                                          decode_query(join_with))
        mediator_logger.info('words for tab_id %s (match_regex %s, join_with %s): %s',
                             tab_id, match_regex, join_with, words)
        return '\n'.join(words)

    def get_text(self):
        delimiter_regex = request.args.get('delimiter_regex', DEFAULT_GET_TEXT_DELIMITER_REGEX)
        replace_with = request.args.get('replace_with', DEFAULT_GET_TEXT_REPLACE_WITH)
        lines = self.remote_api.get_text(decode_query(delimiter_regex),
                                         decode_query(replace_with))
        return '\n'.join(lines)

    def get_html(self):
        delimiter_regex = request.args.get('delimiter_regex', DEFAULT_GET_HTML_DELIMITER_REGEX)
        replace_with = request.args.get('replace_with', DEFAULT_GET_HTML_REPLACE_WITH)
        lines = self.remote_api.get_html(decode_query(delimiter_regex),
                                         decode_query(replace_with))
        return '\n'.join(lines)

    def get_pid(self):
        mediator_logger.info('getting pid')
        return str(os.getpid())

    def get_browser(self):
        mediator_logger.info('getting browser name')
        return self.remote_api.get_browser()

    def echo(self):
        title = request.args.get('title', 'title')
        body = request.args.get('body', 'body')
        reply = ('<html><head><title>%s</title></head>'
                 '<body>%s</body></html>'
                 % (title, body))
        return reply



DEFAULT_GET_WORDS_MATCH_REGEX = r'/\w+/g'
DEFAULT_GET_WORDS_JOIN_WITH = r'"\n"'

DEFAULT_GET_TEXT_DELIMITER_REGEX = r'/\n|\r|\t/g'
DEFAULT_GET_TEXT_REPLACE_WITH = r'" "'

DEFAULT_GET_HTML_DELIMITER_REGEX = r'/\n|\r|\t/g'
DEFAULT_GET_HTML_REPLACE_WITH = r'" "'



import os
from unittest import TestCase
from unittest.mock import patch

from brotab.inout import edit_tabs_in_editor


class TestEditor(TestCase):
    @patch('brotab.inout.run_editor')
    @patch('platform.system', side_effect=['Linux'])
    @patch('os.environ', new={})
    def test_run_editor_linux(self, _system_mock, _run_editor_mock):
        assert ['1'] == edit_tabs_in_editor(['1'])
        editor, filename = _run_editor_mock.call_args[0]
        assert editor == 'nvim'
        assert not os.path.exists(filename)

    @patch('brotab.inout.run_editor')
    @patch('platform.system', side_effect=['Windows'])
    @patch('os.environ', new={})
    def test_run_editor_windows(self, _system_mock, _run_editor_mock):
        assert ['1'] == edit_tabs_in_editor(['1'])
        editor, filename = _run_editor_mock.call_args[0]
        assert editor == 'notepad'
        assert not os.path.exists(filename)

    @patch('brotab.inout.run_editor')
    @patch('platform.system', side_effect=['Windows'])
    @patch('os.environ', new={'EDITOR': 'custom'})
    def test_run_editor_windows_custom(self, _system_mock, _run_editor_mock):
        assert ['1'] == edit_tabs_in_editor(['1'])
        editor, filename = _run_editor_mock.call_args[0]
        assert editor == 'custom'
        assert not os.path.exists(filename)



class Tab:
    def __init__(self, prefix, window_id, tab_id, title, url):
        self.prefix = prefix
        self.window_id = window_id
        self.tab_id = tab_id
        self.title = title
        self.url = url

    @property
    def id(self):
        return '{prefix}.{window_id}.{tab_id}'.format(
            prefix=self.prefix,
            window_id=self.window_id,
            tab_id=self.tab_id,
        )

    @property
    def line(self):
        return '{prefix}.{window_id}.{tab_id}\t{title}\t{url}'.format(
            prefix=self.prefix,
            window_id=self.window_id,
            tab_id=self.tab_id,
            title=self.title,
            url=self.url
        )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash(self.line)

    def __repr__(self):
        return self.line

    @staticmethod
    def from_line(line):
        ids, title, url = line.split('\t')
        prefix, window_id, tab_id = ids.split('.')
        return Tab(prefix, int(window_id), int(tab_id), title, url)


def parse_tab_lines(tab_lines):
    return [Tab.from_line(line) for line in tab_lines]


def iter_window_tabs(left: [Tab], right: [Tab]):
    # get_window_id = attrgetter('window_id')
    # left = sorted(left, key=get_window_id)
    # right = sorted(right, key=get_window_id)
    # groupby(left, get_window_id)

    all_window_ids = set(tab.window_id for tab in left + right)
    for window_id in all_window_ids:
        matching_left = list(filter(lambda x: x.window_id == window_id, left))
        matching_right = list(filter(lambda x: x.window_id == window_id, right))
        yield window_id, matching_left, matching_right



from unittest import TestCase

from brotab.utils import split_tab_ids


class TestUtils(TestCase):
    def test_split_tab_ids(self):
        text = 'c.1.0 c.1.1\tc.1.2\r\nc.1.3 \r\t\n'
        expected = ['c.1.0', 'c.1.1', 'c.1.2', 'c.1.3']
        self.assertEqual(expected, split_tab_ids(text))



from typing import Callable

from brotab.tab import Tab


# def _get_old_index(tab, tabs_before):
#     for index, value in enumerate(tabs_before):
#         if value == tab:
#             return index

#     raise ValueError('Tab %s not found' % tab)


# def _get_tab_id(tab):
#     index = tab.split('\t')[0]
#     str_index = index.split('.')[1]
#     return int(str_index)


def _get_index_by_tab_id(tab_id, tabs: [Tab]):
    for index, tab in enumerate(tabs):
        # if tab_id == _get_tab_id(tab):
        if tab_id == tab.tab_id:
            return index

    return None


class KeyToIndexMapper:
    """
    This class allows building different kind of mappings to retrieve
    values faster later.
    """

    def __init__(self, make_key: Callable, tabs: [Tab]):
        self._make_key = make_key
        self._mapping = {
            make_key(tab): index
            for index, tab in enumerate(tabs)
        }

    def __contains__(self, item):
        key = self._make_key(item)
        return key in self._mapping

    def __getitem__(self, item):
        key = self._make_key(item)
        return self._mapping[key]


class LineToIndexMapper(KeyToIndexMapper):
    def __init__(self, tabs: [Tab]):
        super().__init__(lambda tab: tab.line, tabs)


class TabIdTitleToIndexMapper(KeyToIndexMapper):
    def __init__(self, tabs: [Tab]):
        super().__init__(lambda tab: (tab.tab_id, tab.title), tabs)


# class TabIdIndexUrlToIndexMapper(KeyToIndexMapper):
#     def __init__(self, tabs: [Tab]):
#         super().__init__(lambda tab: (tab.tab_id, tab.title, tab.url), tabs)


def get_longest_increasing_subsequence(X):
    """Returns the Longest Increasing Subsequence in the Given List/Array"""
    N = len(X)
    P = [0] * N
    M = [0] * (N + 1)
    L = 0
    for i in range(N):
        lo = 1
        hi = L
        while lo <= hi:
            mid = (lo + hi) // 2
            if (X[M[mid]] < X[i]):
                lo = mid + 1
            else:
                hi = mid - 1

        newL = lo
        P[i] = M[newL - 1]
        M[newL] = i

        if (newL > L):
            L = newL

    S = []
    k = M[L]
    for i in range(L - 1, -1, -1):
        S.append(X[k])
        k = P[k]
    return S[::-1]


def infer_delete_commands(tabs_before: [Tab], tabs_after: [Tab]):
    tab_id_title_to_index = TabIdTitleToIndexMapper(tabs_after)

    commands = []
    after = set(tabs_after)

    for index in range(len(tabs_before) - 1, -1, -1):
        tab_before = tabs_before[index]

        # if tab_before not in after:
        if tab_before not in tab_id_title_to_index:
            # commands.append(_get_tab_id(tab_before))
            # commands.append(tab_before.tab_id)
            commands.append('%s.%s.%s' % (tab_before.prefix,
                                          tab_before.window_id,
                                          tab_before.tab_id))
    return commands


def _get_old_index(tab_after: Tab,
                   line_to_index: LineToIndexMapper,
                   tab_id_title_to_index: TabIdTitleToIndexMapper,
                   tabs_before: [Tab]):
    """
    Try to find out the index of the tab before the move.
    """
    # Easy case: tab has been just moved without changing window ID.
    if tab_after in line_to_index:
        return line_to_index[tab_after]

    # The tab might have window ID changed. Try to look for the title only.
    if tab_after in tab_id_title_to_index:
        return tab_id_title_to_index[tab_after]

    # print('TAB AFTER', tab_after)
    # print('TAB AFTER LINE', tab_after.line)
    # print('KEY', key)
    #
    # from pprint import pprint
    # print('tab_line_to_old_index')
    # pprint(tab_line_to_old_index)
    #
    # print('tab_id_title_url_to_index')
    # pprint(tab_id_title_url_to_index)
    #
    # print('partilally matching keys in tab_line_to_old_index')
    # for line in tab_line_to_old_index:
    #     if tab_after.title in line:
    #         print('> %s' % line)

    # We're out of ideas...
    raise KeyError('Could not find line: "%s"' % tab_after.line)


def infer_move_commands(tabs_before: [Tab], tabs_after: [Tab]):
    """
    `tabs_before` and `tabs_after` contain an integer in the beginning
    but that's a tab ID, not a position. Thus, a move command means:

        move <tab_id> <to_index>

    where <to_index> is an index within a browser window. Consider this:

    Before:         After:
    f.4\ta          f.8\ta
    f.8\ta          f.4\ta
    f.1\aa          f.1\ta

    The corresponding move commands:

        move f.8 0

    """
    # Remember which tab corresponds to which index in the old list
    line_to_index = LineToIndexMapper(tabs_before)
    # XXX: use tab ID + title as the key
    tab_id_title_to_index = TabIdTitleToIndexMapper(tabs_before)

    # Now see how indices have been reordered by user
    # reordered_indices = [line_to_index[tab.line] for tab in tabs_after]
    reordered_indices = [_get_old_index(tab_after,
                                        line_to_index,
                                        tab_id_title_to_index,
                                        tabs_before)
                         for tab_after in tabs_after]
    # These indices are in correct order, we should not touch them
    correctly_ordered_new_indices = set(
        get_longest_increasing_subsequence(reordered_indices))

    upward, downward = [], []

    # print('reordered_indices', reordered_indices)
    # print('tab_id_title_to_index', tab_id_title_url_to_old_index)

    for new_index, old_index in enumerate(reordered_indices):
        tab_before = tabs_before[old_index]
        tab_after = tabs_after[new_index]

        index_changed = old_index not in correctly_ordered_new_indices
        window_changed = tab_before.window_id != tab_after.window_id

        if index_changed or window_changed:
            triplet = (tab_before.tab_id, tab_after.window_id, new_index)
            upward.append(triplet) if new_index > old_index else downward.append(triplet)
    commands = downward + list(reversed(upward))
    return commands


def make_update(tabId=None,
                active=None,
                autoDiscardable=None,
                highlighted=None,
                muted=None,
                pinned=None,
                url=None,
                openerTabId=None):
    if tabId is None: raise ValueError('tabId is not specified')
    op = {'tab_id': tabId, 'properties': {}}
    if active is not None: op['properties']['active'] = active
    if autoDiscardable is not None: op['properties']['autoDiscardable'] = autoDiscardable
    if highlighted is not None: op['properties']['highlighted'] = highlighted
    if muted is not None: op['properties']['muted'] = muted
    if pinned is not None: op['properties']['pinned'] = pinned
    if url is not None: op['properties']['url'] = url
    if openerTabId is not None: op['properties']['openerTabId'] = openerTabId
    return op


def infer_update_commands(tabs_before: [Tab], tabs_after: [Tab]):
    updates = []
    for tab_before, tab_after in zip(tabs_before, tabs_after):
        if tab_before.url != tab_after.url:
            updates.append(make_update(tabId=tab_after.tab_id, url=tab_after.url))
    return updates


def apply_delete_commands(tabs_before: [Tab], delete_commands):
    tabs = tabs_before[:]
    # for tab_id in delete_commands:
    for delete_command in delete_commands:
        prefix, window_id, tab_id = delete_command.split('.')
        window_id, tab_id = int(window_id), int(tab_id)
        # tab_id = int(command.split()[1])
        del tabs[_get_index_by_tab_id(tab_id, tabs)]
    return tabs


def apply_move_commands(tabs_before: [Tab], move_commands):
    tabs = tabs_before[:]
    for tab_id, window_id, index_to in move_commands:
        index_from = _get_index_by_tab_id(tab_id, tabs)
        tab = tabs.pop(index_from)
        # XXX: should we update the window_id?
        tab.window_id = window_id
        tabs.insert(index_to, tab)
    return tabs


def apply_update_commands(tabs_before: [Tab], update_commands):
    tabs = tabs_before[:]
    for command in update_commands:
        tab_id = command['tab_id']
        index = _get_index_by_tab_id(tab_id, tabs)
        tabs[index].url = command['properties']['url']
    return tabs


def infer_all_commands(tabs_before: [Tab], tabs_after: [Tab]):
    """
    This command takes browser tabs before the edit and after the edit and
    infers a sequence of commands that need to be executed in a browser
    to make transform state from `tabs_before` to `tabs_after`.

    Sample input:
        f.0.0	GMail
        f.0.1	Posix man
        f.0.2	news

    Sample output:
        m 0 5,m 1 1,d 2
    Means:
        move 0 to index 5,
        move 1 to index 1,
        delete 2

    Note that after moves and deletes, indices do not need to be adjusted on the
    browser side. All the indices are calculated by the client program so that
    the JS extension can simply execute the commands without thinking.
    """
    # For now, let's work only within chunks of tabs grouped by
    # the windowId, i.e. moves between windows of the same browser are not
    # supported yet.
    delete_commands, move_commands, update_commands = [], [], []
    # for _window_id, chunk_before, chunk_after in iter_window_tabs(tabs_before, tabs_after):
    #     delete_commands.extend(infer_delete_commands(chunk_before, chunk_after))
    #     chunk_before = apply_delete_commands(chunk_before, delete_commands)
    #     move_commands.extend(infer_move_commands(chunk_before, chunk_after))

    # I've added moves across different existing (!) windows of the same
    # browser to iter_window_tabs is not used.

    # [_] detect a move to a new nonexistent window

    delete_commands.extend(infer_delete_commands(tabs_before, tabs_after))
    tabs_before = apply_delete_commands(tabs_before, delete_commands)
    move_commands.extend(infer_move_commands(tabs_before, tabs_after))
    tabs_before = apply_move_commands(tabs_before, move_commands)
    update_commands.extend(infer_update_commands(tabs_before, tabs_after))

    return delete_commands, move_commands, update_commands



import os
import signal
import time
from multiprocessing import Process
from threading import Thread
from typing import Callable

# from psutil import pid_exists

from brotab.mediator.log import disable_logging
from brotab.mediator.log import mediator_logger


class NotStarted(Exception):
    pass


class Runner:
    def __init__(self, serve: Callable[[], None], shutdown: Callable[[bool], None]):
        self._serve = serve
        self._shutdown = shutdown

    def shutdown(self, join: bool) -> None:
        # TODO: break this to test ctrl-c
        if not self._shutdown:
            raise NotStarted('start the runner first')
        mediator_logger.info('Runner: calling terminate (pid=%s): %s', os.getpid(), self._shutdown)
        self._shutdown(join)

    def here(self) -> None:
        mediator_logger.info('Started mediator process, pid=%s', os.getpid())
        disable_logging()
        return self._serve()

    def in_thread(self) -> Thread:
        thread = Thread(target=self.here)
        thread.daemon = True
        thread.start()
        return thread

    # def parent_watcher(self, running: Callable, interval: float):
    #     self._watcher(running, os.getppid(), interval=interval)
    #
    # def _watcher(self, running: Callable, parent_pid: int, interval: float) -> None:
    #     mediator_logger.info('Watching parent process parent=%s current pid=%s',
    #                          parent_pid, os.getpid())
    #     while True:
    #         time.sleep(interval)
    #         if not running():  # someone shutdown mediator, let's bail
    #             break
    #         if not pid_exists(parent_pid):
    #             mediator_logger.info('Parent process died pid=%s, shutting down mediator', parent_pid)
    #             self.shutdown(join=False)
    #             break



#!/usr/bin/env python3

"""
This is a browser tab client. It allows listing, closing and creating
tabs in browsers from command line. Currently Firefox and Chrome are
supported.

To enable RPC in Chrome, run it as follows:

    chromium-browser --remote-debugging-port=9222 &!

To enable RPC in Firefox, install Mozrepl plugin:

    https://addons.mozilla.org/en-US/firefox/addon/mozrepl/
    https://github.com/bard/mozrepl/wiki

    Change port to 4242, and tick Tools -> MozRepl -> Activate on startup

Todo:
    [_] add rt-browsers support for Chromium (grab tabs from from database)
    [_] add rt-browsers-history to grab rt tickets from browser history

News:

    Starting from Firefox 55 mozrepl is not working anymore. Even worse, they
    guarantee that since Firefox 57 extensions that have not transitioned to
    WebExtensions technology will stop working. I need to find a replacement
    for mozrepl. As I need only a limited set of its potential functionality,
    implementing my own extension sounds like a viable idea. Two things are
    required:

        0. extensions setup basics:

        https://developer.mozilla.org/en-US/Add-ons/WebExtensions/What_are_WebExtensions

        1. tabs api:

        https://developer.mozilla.org/en-US/Add-ons/WebExtensions/API/tabs
        https://developer.mozilla.org/en-US/Add-ons/WebExtensions/API/tabs/query
        https://github.com/mdn/webextensions-examples/blob/master/tabs-tabs-tabs/tabs.js

        2. native messaging:

        https://developer.mozilla.org/en-US/Add-ons/WebExtensions/Native_messaging
        https://developer.mozilla.org/en-US/Add-ons/WebExtensions/manifest.json/permissions
        https://github.com/mdn/webextensions-examples/tree/master/native-messaging

"""

import os
import re
import sys
import time
from argparse import ArgumentParser
from functools import partial
from itertools import groupby
from json import loads
from string import ascii_lowercase
from typing import List
from typing import Tuple
from urllib.parse import quote_plus

from brotab.api import MultipleMediatorsAPI
from brotab.api import SingleMediatorAPI
from brotab.const import DEFAULT_GET_HTML_DELIMITER_REGEX
from brotab.const import DEFAULT_GET_HTML_REPLACE_WITH
from brotab.const import DEFAULT_GET_TEXT_DELIMITER_REGEX
from brotab.const import DEFAULT_GET_TEXT_REPLACE_WITH
from brotab.const import DEFAULT_GET_WORDS_JOIN_WITH
from brotab.const import DEFAULT_GET_WORDS_MATCH_REGEX
from brotab.inout import get_mediator_ports
from brotab.inout import in_temp_dir
from brotab.inout import is_port_accepting_connections
from brotab.inout import marshal
from brotab.inout import read_stdin
from brotab.inout import read_stdin_lines
from brotab.inout import stdout_buffer_write
from brotab.mediator.log import brotab_logger
from brotab.operations import make_update
from brotab.platform import is_windows
from brotab.platform import make_windows_path_double_sep
from brotab.platform import register_native_manifest_windows_brave
from brotab.platform import register_native_manifest_windows_chrome
from brotab.platform import register_native_manifest_windows_firefox
from brotab.search.index import index
from brotab.search.query import query
from brotab.utils import get_file_size
from brotab.utils import split_tab_ids
from brotab.utils import which


def parse_target_hosts(target_hosts: str) -> Tuple[List[str], List[int]]:
    """
    Input: localhost:2000,127.0.0.1:3000
    Output: (['localhost', '127.0.0.1'], [2000, 3000])
    """
    hosts, ports = [], []
    for pair in target_hosts.split(','):
        host, port = pair.split(':')
        hosts.append(host)
        ports.append(int(port))
    return hosts, ports


def create_clients(target_hosts=None) -> List[SingleMediatorAPI]:
    if target_hosts is None:
        ports = list(get_mediator_ports())
        hosts = ['localhost'] * len(ports)
    else:
        hosts, ports = parse_target_hosts(target_hosts)

    result = [SingleMediatorAPI(prefix, host=host, port=port)
              for prefix, host, port in zip(ascii_lowercase, hosts, ports)
              if is_port_accepting_connections(port, host)]
    brotab_logger.info('Created clients: %s', result)
    return result


def parse_prefix_and_window_id(prefix_window_id):
    prefix, window_id = None, None
    try:
        prefix, window_id = prefix_window_id.split('.')
        prefix += '.'
        window_id = window_id or None
    except ValueError:
        prefix = prefix_window_id
        prefix += '' if prefix.endswith('.') else '.'

    return prefix, window_id


def move_tabs(args):
    brotab_logger.info('Moving tabs')
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    api.move_tabs([])


def list_tabs(args):
    """
    Use this to show duplicates:
        bt list | sort -k3 | uniq -f2 -D | cut -f1 | bt close
    """
    brotab_logger.info('Listing tabs')
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    tabs = api.list_tabs([])
    message = '\n'.join(tabs) + '\n'
    sys.stdout.buffer.write(message.encode('utf8'))


def close_tabs(args):
    # Try stdin if arguments are empty
    tab_ids = args.tab_ids
    if len(args.tab_ids) == 0:
        tab_ids = split_tab_ids(read_stdin().strip())

    brotab_logger.info('Closing tabs: %s', tab_ids)
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    tabs = api.close_tabs(tab_ids)


def activate_tab(args):
    brotab_logger.info('Activating tab: %s', args.tab_id)
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    api.activate_tab(args.tab_id, args.focused)


def show_active_tabs(args):
    brotab_logger.info('Showing active tabs: %s', args)
    apis = create_clients(args.target_hosts)
    for api in apis:
        tabs = api.get_active_tabs(args)
        for tab in tabs:
            print('%s\t%s' % (tab, api))


def search_tabs(args):
    for result in query(args.sqlite, args.query):
        print('\t'.join([result.tab_id, result.title, result.snippet]))


def query_tabs(args):
    brotab_logger.info('Querying tabs: %s', args)
    d = vars(args)
    if d['info'] is not None:
        queryInfo = d['info']
    else:
        queryInfo = {k: v for k, v in d.items()
                     if v is not None and k not in ['func', 'info', 'target_hosts']}
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    for tab in api.query_tabs(queryInfo):
        print(tab)


def index_tabs(args):
    if args.tsv is None:
        args.tsv = in_temp_dir('tabs.tsv')
        args.cleanup = True
        brotab_logger.info(
            'index_tabs: retrieving tabs from browser into file %s', args.tsv)
        start = time.time()
        get_text(args)
        delta = time.time() - start
        brotab_logger.info('getting text took %s', delta)

    start = time.time()
    index(args.sqlite, args.tsv)
    delta = time.time() - start
    brotab_logger.info('sqlite create took %s, size %s',
                       delta, get_file_size(args.sqlite))


def new_tab(args):
    prefix, window_id = parse_prefix_and_window_id(args.prefix_window_id)
    search_query = ' '.join(args.query)
    brotab_logger.info('Opening search for "%s", prefix "%s", window_id "%s"',
                       search_query, prefix, window_id)
    url = "https://www.google.com/search?q=%s" % quote_plus(search_query)
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    ids = api.open_urls([url], prefix, window_id)
    stdout_buffer_write(marshal(ids))


def open_urls(args):
    """
    curl -X POST 'http://localhost:4626/open_urls' -F 'urls=@urls.txt'
    curl -X POST 'http://localhost:4627/open_urls' -F 'urls=@urls.txt' -F 'window_id=749'

    where urls.txt contains one url per line (not JSON)
    """
    prefix, window_id = parse_prefix_and_window_id(args.prefix_window_id)
    urls = read_stdin_lines()
    brotab_logger.info('Opening URLs from stdin, prefix "%s", window_id "%s": %s',
                       prefix, window_id, urls)
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    ids = api.open_urls(urls, prefix, window_id)
    stdout_buffer_write(marshal(ids))


def navigate_urls(args):
    """
    curl -X POST 'http://localhost:4626/update_tabs' --data '{"tab_id": 20, "properties": { "url": "https://www.google.com" }}'
    """
    raw = read_stdin(timeout=0.05)
    if raw:
        pairs = [x.strip().split('\t') for x in raw.splitlines()]
        updates = [make_update(tabId=tab_id, url=url) for tab_id, url in pairs]
    else:
        updates = [make_update(tabId=args.tab_id, url=args.url)]
    brotab_logger.info('Navigating: %s', updates)
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    results = api.update_tabs(updates)
    stdout_buffer_write(marshal(results))


def update_tabs(args):
    """
    curl -X POST 'http://localhost:4626/update_tabs' --data '{"tab_id": 20, "properties": { "url": "https://www.google.com" }}'
    """
    raw = read_stdin(timeout=0.01).strip()
    if raw:
        updates = loads(raw)
    else:
        d = vars(args)
        if d['info'] is not None:
            updates = [d['info']]
        else:
            updates = {k: v for k, v in d.items()
                       if v is not None and k not in ['func', 'info', 'target_hosts']}
            if 'tabId' not in updates: raise ValueError('tabId is required')
            updates = [make_update(**updates)]
    brotab_logger.info('Updating tabs: %s', updates)
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    results = api.update_tabs(updates)
    stdout_buffer_write(marshal(results))


def get_words(args):
    # return tab.execute({javascript: "
    # [...new Set(document.body.innerText.match(/\w+/g))].sort().join('\n');
    # "})
    start = time.time()
    brotab_logger.info('Get words from tabs: %s, match_regex=%s, join_with=%s',
                       args.tab_ids, args.match_regex, args.join_with)
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    words = api.get_words(args.tab_ids, args.match_regex, args.join_with)
    print('\n'.join(words))
    delta = time.time() - start
    # print('DELTA TOTAL', delta, file=sys.stderr)


def get_text_or_html(getter, args):
    tabs = getter([], args.delimiter_regex, args.replace_with)
    re_match_tabs = re.compile('|'.join(['^%s\t' % tab for tab in args.tab_ids]))
    tabs = [tab for tab in tabs if re_match_tabs.match(tab)]

    if args.cleanup:
        pattern = re.compile(r'\s+')
        old_tabs = tabs
        tabs = []
        for line in old_tabs:
            tab_id, title, url, text = line.split('\t')
            text = re.sub(pattern, ' ', text)
            tabs.append('\t'.join([tab_id, title, url, text]))

    message = '\n'.join(tabs) + '\n'
    if args.tsv is None:
        stdout_buffer_write(message.encode('utf8'))
    else:
        with open(args.tsv, 'w', encoding='utf-8') as file_:
            file_.write(message)


def get_text(args):
    brotab_logger.info('Get text from tabs')
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    return get_text_or_html(api.get_text, args)


def get_html(args):
    brotab_logger.info('Get html from tabs')
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    return get_text_or_html(api.get_html, args)


def show_duplicates(args):
    # I'm not using uniq here because it's not easy to get duplicates
    # only by a single column. awk is much easier in this regard.
    # print('bt list | sort -k3 | uniq -f2 -D | cut -f1 | bt close')
    print("Show duplicates by Title:")
    print(
        "bt list | sort -k2 | awk -F$'\\t' '{ if (a[$2]++ > 0) print }' | cut -f1 | bt close")
    print("")
    print("Show duplicates by URL:")
    print(
        "bt list | sort -k3 | awk -F$'\\t' '{ if (a[$3]++ > 0) print }' | cut -f1 | bt close")


def _get_window_id(tab):
    ids, _title, _url = tab.split('\t')
    client_id, window_id, tab_id = ids.split('.')
    return '%s.%s' % (client_id, window_id)


def _print_available_windows(tabs):
    for key, group in groupby(sorted(tabs), _get_window_id):
        group = list(group)
        print('%s\t%s' % (key, len(group)))


def show_windows(args):
    brotab_logger.info('Showing windows')
    api = MultipleMediatorsAPI(create_clients(args.target_hosts))
    tabs = api.list_tabs([])
    _print_available_windows(tabs)


def show_clients(args):
    brotab_logger.info('Showing clients')
    for client in create_clients(args.target_hosts):
        print(client)


def install_mediator(args):
    brotab_logger.info('Installing mediators')
    bt_mediator_path = which('bt_mediator')
    if is_windows():
        bt_mediator_path = make_windows_path_double_sep(bt_mediator_path)

    native_app_manifests = [
        ('mediator/firefox_mediator.json',
         '~/.mozilla/native-messaging-hosts/brotab_mediator.json'),
        ('mediator/chromium_mediator.json',
         '~/.config/chromium/NativeMessagingHosts/brotab_mediator.json'),
        ('mediator/chromium_mediator.json',
         '~/.config/google-chrome/NativeMessagingHosts/brotab_mediator.json'),
        ('mediator/chromium_mediator.json',
         '~/.config/BraveSoftware/Brave-Browser/NativeMessagingHosts/brotab_mediator.json'),
    ]

    if args.tests:
        native_app_manifests.append(
            ('mediator/chromium_mediator_tests.json',
             '~/.config/chromium/NativeMessagingHosts/brotab_mediator.json'))
        native_app_manifests.append(
            ('mediator/chromium_mediator_tests.json',
             '~/.config/google-chrome/NativeMessagingHosts/brotab_mediator.json'))

    from pkg_resources import resource_string
    for filename, destination in native_app_manifests:
        destination = os.path.expanduser(os.path.expandvars(destination))
        template = resource_string(__name__, filename).decode('utf8')
        manifest = template.replace(r'$PWD/brotab_mediator.py', bt_mediator_path)
        brotab_logger.info('Installing template %s into %s', filename, destination)
        print('Installing mediator manifest %s' % destination)

        os.makedirs(os.path.dirname(destination), exist_ok=True)
        with open(destination, 'w') as file_:
            file_.write(manifest)

        if is_windows() and 'mozilla' in destination:
            register_native_manifest_windows_firefox(destination)
        if is_windows() and 'chrome' in destination:
            register_native_manifest_windows_chrome(destination)
        if is_windows() and 'Brave' in destination:
            register_native_manifest_windows_brave(destination)

    print('Link to Firefox extension: https://addons.mozilla.org/en-US/firefox/addon/brotab/')
    print(
        'Link to Chrome (Chromium)/Brave extension: https://chrome.google.com/webstore/detail/brotab/mhpeahbikehnfkfnmopaigggliclhmnc/')


def executejs(args):
    pass


def no_command(parser, args):
    print('No command has been specified')
    parser.print_help()
    return 1


def parse_args(args):
    parser = ArgumentParser(
        description='''
        bt (brotab = Browser Tabs) is a command-line tool that helps you manage
        browser tabs. It can help you list, close, reorder, open and activate
        your tabs.
        ''')

    parser.add_argument('--target', dest='target_hosts', default=None,
                        help='Target hosts IP:Port')

    subparsers = parser.add_subparsers()
    parser.set_defaults(func=partial(no_command, parser))

    parser_move_tabs = subparsers.add_parser(
        'move',
        help='''
        move tabs around. This command lists available tabs and runs
        the editor. In the editor you can 1) reorder tabs -- tabs will
        be moved in the browser 2) delete tabs -- tabs will be closed
        3) change window ID of the tabs -- tabs will be moved to
        specified windows
        ''')
    parser_move_tabs.set_defaults(func=move_tabs)

    parser_list_tabs = subparsers.add_parser(
        'list',
        help='''
        list available tabs. The command will request all available clients
        (browser plugins, mediators), and will display browser tabs in the
        following format:
        "<prefix>.<window_id>.<tab_id><Tab>Page title<Tab>URL"
        ''')
    parser_list_tabs.set_defaults(func=list_tabs)

    parser_close_tabs = subparsers.add_parser(
        'close',
        help='''
        close specified tab IDs. Tab IDs should be in the following format:
        "<prefix>.<window_id>.<tab_id>". You can use "list" command to obtain
        tab IDs (first column)
        ''')
    parser_close_tabs.set_defaults(func=close_tabs)
    parser_close_tabs.add_argument('tab_ids', type=str, nargs='*',
                                   help='Tab IDs to close')

    parser_activate_tab = subparsers.add_parser(
        'activate',
        help='''
        activate given tab ID. Tab ID should be in the following format:
        "<prefix>.<window_id>.<tab_id>"
        ''')
    parser_activate_tab.set_defaults(func=activate_tab)
    parser_activate_tab.add_argument('tab_id', type=str, nargs=1,
                                     help='Tab ID to activate')
    parser_activate_tab.add_argument('--focused', action='store_const', const=True, default=None,
                                     help='make browser focused after tab activation (default: False)')

    parser_active_tab = subparsers.add_parser(
        'active',
        help='''
        display active tab for each client/window in the following format:
        "<prefix>.<window_id>.<tab_id>"
        ''')
    parser_active_tab.set_defaults(func=show_active_tabs)

    parser_search_tabs = subparsers.add_parser(
        'search',
        help='''
        Search across your indexed tabs using sqlite fts5 plugin.
        ''')
    parser_search_tabs.set_defaults(func=search_tabs)
    parser_search_tabs.add_argument('--sqlite', type=str, default=in_temp_dir('tabs.sqlite'),
                                    help='sqlite DB filename')
    parser_search_tabs.add_argument('query', type=str, help='Search query')

    parser_query_tabs = subparsers.add_parser(
        'query',
        help='Filter tabs using chrome.tabs api.',
        prefix_chars='-+')
    parser_query_tabs.set_defaults(func=query_tabs)
    parser_query_tabs.add_argument('+active', action='store_const', const=True, default=None,
                                   help='tabs are active in their windows')
    parser_query_tabs.add_argument('-active', action='store_const', const=False, default=None,
                                   help='tabs are not active in their windows')
    parser_query_tabs.add_argument('+pinned', action='store_const', const=True, default=None,
                                   help='tabs are pinned')
    parser_query_tabs.add_argument('-pinned', action='store_const', const=False, default=None,
                                   help='tabs are not pinned')
    parser_query_tabs.add_argument('+audible', action='store_const', const=True, default=None,
                                   help='tabs are audible')
    parser_query_tabs.add_argument('-audible', action='store_const', const=False, default=None,
                                   help='tabs are not audible')
    parser_query_tabs.add_argument('+muted', action='store_const', const=True, default=None,
                                   help='tabs are muted')
    parser_query_tabs.add_argument('-muted', action='store_const', const=False, default=None,
                                   help='tabs not are muted')
    parser_query_tabs.add_argument('+highlighted', action='store_const', const=True, default=None,
                                   help='tabs are highlighted')
    parser_query_tabs.add_argument('-highlighted', action='store_const', const=False, default=None,
                                   help='tabs not are highlighted')
    parser_query_tabs.add_argument('+discarded', action='store_const', const=True, default=None,
                                   help='tabs are discarded i.e. unloaded from memory but still visible in the tab strip.')
    parser_query_tabs.add_argument('-discarded', action='store_const', const=False, default=None,
                                   help='tabs are not discarded i.e. unloaded from memory but still visible in the tab strip.')
    parser_query_tabs.add_argument('+autoDiscardable', action='store_const', const=True, default=None,
                                   help='tabs can be discarded automatically by the browser when resources are low.')
    parser_query_tabs.add_argument('-autoDiscardable', action='store_const', const=False, default=None,
                                   help='tabs cannot be discarded automatically by the browser when resources are low.')
    parser_query_tabs.add_argument('+currentWindow', action='store_const', const=True, default=None,
                                   help='tabs are in the current window.')
    parser_query_tabs.add_argument('-currentWindow', action='store_const', const=False, default=None,
                                   help='tabs are not in the current window.')
    parser_query_tabs.add_argument('+lastFocusedWindow', action='store_const', const=True, default=None,
                                   help='tabs are in the last focused window.')
    parser_query_tabs.add_argument('-lastFocusedWindow', action='store_const', const=False, default=None,
                                   help='tabs are not in the last focused window.')
    parser_query_tabs.add_argument('-status', type=str, choices=['loading', 'complete'],
                                   help='whether the tabs have completed loading i.e. loading or complete.')
    parser_query_tabs.add_argument('-title', type=str,
                                   help='match page titles against a pattern.')
    parser_query_tabs.add_argument('-url', type=str, action='append',
                                   help='match tabs against one or more URL patterns. Fragment identifiers are not matched. see https://developer.chrome.com/extensions/match_patterns')
    parser_query_tabs.add_argument('-windowId', type=int,
                                   help='the ID of the parent window, or windows.WINDOW_ID_CURRENT for the current window.')
    parser_query_tabs.add_argument('-windowType', type=str, choices=['normal', 'popup', 'panel', 'app', 'devtools'],
                                   help='the type of window the tabs are in.')
    parser_query_tabs.add_argument('-index', type=int,
                                   help='the position of the tabs within their windows.')
    parser_query_tabs.add_argument('-info', type=str,
                                   help='''
        the queryInfo parameter as outlined here: https://developer.chrome.com/extensions/tabs#method-query.
        all other query arguments are ignored if this argument is present.
        ''')

    parser_index_tabs = subparsers.add_parser(
        'index',
        help='''
        Index the text from browser's tabs. Text is put into sqlite fts5 table.
        ''')
    parser_index_tabs.set_defaults(func=index_tabs)
    parser_index_tabs.add_argument('tab_ids', type=str, nargs='*',
                                   help='Tab IDs to get text from')
    parser_index_tabs.add_argument('--sqlite', type=str, default=in_temp_dir('tabs.sqlite'),
                                   help='sqlite DB filename')
    parser_index_tabs.add_argument('--tsv', type=str, default=None,
                                   help='get text from tabs and index the results')
    parser_index_tabs.add_argument(
        '--delimiter-regex', type=str, default=DEFAULT_GET_TEXT_DELIMITER_REGEX,
        help='Regex that is used to match delimiters in the page text')
    parser_index_tabs.add_argument(
        '--replace-with', type=str, default=DEFAULT_GET_TEXT_REPLACE_WITH,
        help='String that is used to replaced matched delimiters')

    parser_new_tab = subparsers.add_parser(
        'new',
        help='''
        open new tab with the Google search results of the arguments that
        follow. One positional argument is required:
        <prefix>.<window_id> OR <client>. If window_id is not specified,
        URL will be opened in the active window of the specifed client
        ''')
    parser_new_tab.set_defaults(func=new_tab)
    parser_new_tab.add_argument(
        'prefix_window_id', type=str,
        help='Client prefix and (optionally) window id, e.g. b.20')
    parser_new_tab.add_argument('query', type=str, nargs='*',
                                help='Query to search for in Google')

    parser_open_urls = subparsers.add_parser(
        'open',
        help='''
        open URLs from the stdin (one URL per line). One positional argument is
        required: <prefix>.<window_id> OR <client>. If window_id is not
        specified, URL will be opened in the active window of the specifed
        client
        ''')
    parser_open_urls.set_defaults(func=open_urls)
    parser_open_urls.add_argument(
        'prefix_window_id', type=str,
        help='Client prefix and (optionally) window id, e.g. b.20')

    parser_navigate_urls = subparsers.add_parser(
        'navigate',
        help='''
        navigate to URLs. There are two ways to specify tab ids and URLs:
        1. stdin: lines with pairs of "tab_id<tab>url"
        2. arguments: bt navigate <tab_id> "<url>", e.g. bt navigate b.20.1 "https://google.com"
        stdin has the priority.
        ''')
    parser_navigate_urls.set_defaults(func=navigate_urls)
    parser_navigate_urls.add_argument('tab_id', type=str, help='Tab id e.g. b.20.130')
    parser_navigate_urls.add_argument('url', type=str, help='URL to navigate to')

    parser_update_tabs = subparsers.add_parser(
        'update',
        help='''
        Update tabs state, e.g. URL. There are two ways to specify updates:
        1. stdin, pass JSON of the form:
        [{"tab_id": "b.20.130", "properties": {"url": "http://www.google.com"}}]
        Where "properties" can be anything defined here:
        https://developer.mozilla.org/en-US/docs/Mozilla/Add-ons/WebExtensions/API/tabs/update
        Example:
        echo '[{"tab_id":"a.2118.2156", "properties":{"url":"https://google.com"}}]' | bt update
        
        2. arguments, e.g.: bt update -tabId b.1.862 -url="http://www.google.com" +muted
        ''',
        prefix_chars='-+')
    parser_update_tabs.set_defaults(func=update_tabs)
    parser_update_tabs.add_argument('-tabId', type=str,
                                    help='tab id to apply updates to')
    parser_update_tabs.add_argument('-url', type=str,
                                    help='a URL to navigate the tab to. JavaScript URLs are not supported')
    parser_update_tabs.add_argument('-openerTabId', type=str,
                                    help='the ID of the tab that opened this tab. If specified, the opener tab must be in the same window as this tab')
    parser_update_tabs.add_argument('+active', action='store_const', const=True, default=None,
                                    help='make tab active')
    parser_update_tabs.add_argument('-active', action='store_const', const=False, default=None,
                                    help='does nothing')
    parser_update_tabs.add_argument('+autoDiscardable', action='store_const', const=True, default=None,
                                    help='whether the tab should be discarded automatically by the browser when resources are low')
    parser_update_tabs.add_argument('-autoDiscardable', action='store_const', const=False, default=None,
                                    help='whether the tab should be discarded automatically by the browser when resources are low')
    parser_update_tabs.add_argument('+highlighted', action='store_const', const=True, default=None,
                                    help='adds the tab to the current selection')
    parser_update_tabs.add_argument('-highlighted', action='store_const', const=False, default=None,
                                    help='removes the tab from the current selection')
    parser_update_tabs.add_argument('+muted', action='store_const', const=True, default=None,
                                    help='mute tab')
    parser_update_tabs.add_argument('-muted', action='store_const', const=False, default=None,
                                    help='unmute tab')
    parser_update_tabs.add_argument('+pinned', action='store_const', const=True, default=None,
                                    help='pin tab')
    parser_update_tabs.add_argument('-pinned', action='store_const', const=False, default=None,
                                    help='unpin tab')
    parser_update_tabs.add_argument('-info', type=str,
                                    help='''
        JSON in the following format:
        $ bt update -info '[{"tab_id": "b.20.130", "properties": {"url": "http://www.google.com"}}]'
        all other update arguments are ignored if this argument is present.
        ''')

    parser_get_words = subparsers.add_parser(
        'words',
        help='''
        show sorted unique words from all active tabs of all clients. This is
        a helper for webcomplete plugin that helps complete words from the
        browser
        ''')
    parser_get_words.set_defaults(func=get_words)
    parser_get_words.add_argument('tab_ids', type=str, nargs='*',
                                  help='Tab IDs to get words from')
    parser_get_words.add_argument(
        '--match-regex', type=str, default=DEFAULT_GET_WORDS_MATCH_REGEX,
        help='Regex that is used to match words in the page text')
    parser_get_words.add_argument(
        '--join-with', type=str, default=DEFAULT_GET_WORDS_JOIN_WITH,
        help='String that is used to join matched words')

    parser_get_text = subparsers.add_parser(
        'text',
        help='''
        show text form all tabs
        ''')
    parser_get_text.set_defaults(func=get_text)
    parser_get_text.add_argument('tab_ids', type=str, nargs='*',
                                 help='Tab IDs to get text from')
    parser_get_text.add_argument('--tsv', type=str, default=None,
                                 help='tsv file to save results to')
    parser_get_text.add_argument('--cleanup', action='store_true',
                                 default=False,
                                 help='force removal of extra whitespace')
    parser_get_text.add_argument(
        '--delimiter-regex', type=str, default=DEFAULT_GET_TEXT_DELIMITER_REGEX,
        help='Regex that is used to match delimiters in the page text')
    parser_get_text.add_argument(
        '--replace-with', type=str, default=DEFAULT_GET_TEXT_REPLACE_WITH,
        help='String that is used to replaced matched delimiters')

    parser_get_html = subparsers.add_parser(
        'html',
        help='''
        show html form all tabs
        ''')
    parser_get_html.set_defaults(func=get_html)
    parser_get_html.add_argument('tab_ids', type=str, nargs='*',
                                 help='Tab IDs to get text from')
    parser_get_html.add_argument('--tsv', type=str, default=None,
                                 help='tsv file to save results to')
    parser_get_html.add_argument('--cleanup', action='store_true',
                                 default=False,
                                 help='force removal of extra whitespace')
    parser_get_html.add_argument(
        '--delimiter-regex', type=str, default=DEFAULT_GET_HTML_DELIMITER_REGEX,
        help='Regex that is used to match delimiters in the page text')
    parser_get_html.add_argument(
        '--replace-with', type=str, default=DEFAULT_GET_HTML_REPLACE_WITH,
        help='String that is used to replaced matched delimiters')

    parser_show_duplicates = subparsers.add_parser(
        'dup',
        help='''
        display reminder on how to show duplicate tabs using command-line tools
        ''')
    parser_show_duplicates.set_defaults(func=show_duplicates)

    parser_show_windows = subparsers.add_parser(
        'windows',
        help='''
        display available prefixes and window IDs, along with the number of
        tabs in every window
        ''')
    parser_show_windows.set_defaults(func=show_windows)

    parser_show_clients = subparsers.add_parser(
        'clients',
        help='''
        display available browser clients (mediators), their prefixes and
        address (host:port), native app PIDs, and browser names
        ''')
    parser_show_clients.set_defaults(func=show_clients)

    parser_install_mediator = subparsers.add_parser(
        'install',
        help='''
        configure browser settings to use bt mediator (native messaging app)
        ''')
    parser_install_mediator.add_argument('--tests', action='store_true',
                                         default=False,
                                         help='install testing version of '
                                              'manifest for chromium')
    parser_install_mediator.set_defaults(func=install_mediator)

    return parser.parse_args(args)


def run_commands(args):
    args = parse_args(args)
    result = 0
    try:
        result = args.func(args)
    except BrokenPipeError:
        pass
    return result


def main():
    exit(run_commands(sys.argv[1:]))


if __name__ == '__main__':
    main()



from brotab.const import DEFAULT_GET_HTML_DELIMITER_REGEX
from brotab.const import DEFAULT_GET_HTML_REPLACE_WITH
from brotab.const import DEFAULT_GET_TEXT_DELIMITER_REGEX
from brotab.const import DEFAULT_GET_TEXT_REPLACE_WITH
from brotab.const import DEFAULT_GET_WORDS_JOIN_WITH
from brotab.const import DEFAULT_GET_WORDS_MATCH_REGEX
from brotab.utils import encode_query

DEFAULT_TRANSPORT_TIMEOUT = 60.0
DEFAULT_SHUTDOWN_POLL_INTERVAL = 1.0
DEFAULT_HTTP_IFACE = '127.0.0.1'
DEFAULT_MIN_HTTP_PORT = 4625
DEFAULT_MAX_HTTP_PORT = DEFAULT_MIN_HTTP_PORT + 10
DEFAULT_GET_WORDS_MATCH_REGEX = encode_query(DEFAULT_GET_WORDS_MATCH_REGEX)
DEFAULT_GET_WORDS_JOIN_WITH = encode_query(DEFAULT_GET_WORDS_JOIN_WITH)
DEFAULT_GET_TEXT_DELIMITER_REGEX = encode_query(DEFAULT_GET_TEXT_DELIMITER_REGEX)
DEFAULT_GET_TEXT_REPLACE_WITH = encode_query(DEFAULT_GET_TEXT_REPLACE_WITH)
DEFAULT_GET_HTML_DELIMITER_REGEX = encode_query(DEFAULT_GET_HTML_DELIMITER_REGEX)
DEFAULT_GET_HTML_REPLACE_WITH = encode_query(DEFAULT_GET_HTML_REPLACE_WITH)



import io
import mimetypes
import os
import socket
import sys
import tempfile
import uuid
from subprocess import CalledProcessError
from subprocess import check_call
from tempfile import NamedTemporaryFile
from typing import BinaryIO
from typing import Iterable
from typing import Union

from select import select

from brotab.platform import get_editor

MIN_MEDIATOR_PORT = 4625
MAX_MEDIATOR_PORT = MIN_MEDIATOR_PORT + 10


def slurp(filename):
    with open(filename) as file_:
        return file_.read()


def slurp_lines(filename):
    with open(filename) as file_:
        return [line.strip() for line in file_.readlines()]


def spit(filename, contents):
    with open(filename, 'w', encoding='utf-8') as file_:
        file_.write(contents)


def in_temp_dir(filename) -> str:
    temp_dir = tempfile.gettempdir()
    return os.path.join(temp_dir, filename)


def get_mediator_ports() -> Iterable:
    return range(MIN_MEDIATOR_PORT, MAX_MEDIATOR_PORT)


def get_available_tcp_port(start=1025, end=65536, host='127.0.0.1'):
    for port in range(start, end):
        if not is_port_accepting_connections(port, host):
            return port
    return RuntimeError('Cannot find available port in range %d:%d' % (start, end))


def is_port_accepting_connections(port, host='127.0.0.1'):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(0.100)
    result = s.connect_ex((host, port))
    s.close()
    return result == 0


def save_tabs_to_file(tabs, filename):
    with open(filename, 'w', encoding='utf-8') as file_:
        file_.write('\n'.join(tabs))


def maybe_remove_file(filename):
    if os.path.exists(filename):
        os.remove(filename)


def load_tabs_from_file(filename):
    with open(filename, encoding='utf-8') as file_:
        return [line.strip() for line in file_.readlines()]


def run_editor(executable: str, filename: str):
    return check_call([executable, filename])


def edit_tabs_in_editor(tabs_before):
    with NamedTemporaryFile() as file_:
        file_name = file_.name
        file_.close()
        save_tabs_to_file(tabs_before, file_name)
        try:
            run_editor(get_editor(), file_name)
            tabs_after = load_tabs_from_file(file_name)
            maybe_remove_file(file_name)
            return tabs_after
        except CalledProcessError:
            return None


def read_stdin(timeout=1.0):
    if select([sys.stdin, ], [], [], timeout)[0]:
        return sys.stdin.read()
    return ''


def read_stdin_lines():
    return [line.strip() for line in sys.stdin.readlines()]


def marshal(obj):
    if isinstance(obj, str):
        return obj.encode('utf-8')
    if isinstance(obj, list):
        data = '\n'.join(obj) + '\n'
        return data.encode('utf-8')
    return str(obj).encode('utf-8')


def stdout_buffer_write(message):
    return sys.stdout.buffer.write(message)


# Taken from https://pymotw.com/3/urllib.request/
class MultiPartForm:
    """Accumulate the data to be used when posting a form."""

    def __init__(self):
        self.form_fields = []
        self.files = []
        # Use a large random byte string to separate
        # parts of the MIME data.
        self.boundary = uuid.uuid4().hex.encode('utf-8')
        return

    def get_content_type(self):
        return 'multipart/form-data; boundary={}'.format(
            self.boundary.decode('utf-8'))

    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))

    def add_file(self, fieldname, filename, fileHandle,
                 mimetype=None):
        """Add a file to be uploaded."""
        body = fileHandle.read()
        if mimetype is None:
            mimetype = (
                    mimetypes.guess_type(filename)[0] or
                    'application/octet-stream'
            )
        self.files.append((fieldname, filename, mimetype, body))
        return

    @staticmethod
    def _form_data(name):
        return ('Content-Disposition: form-data; '
                'name="{}"\r\n').format(name).encode('utf-8')

    @staticmethod
    def _attached_file(name, filename):
        return ('Content-Disposition: file; '
                'name="{}"; filename="{}"\r\n').format(
            name, filename).encode('utf-8')

    @staticmethod
    def _content_type(ct):
        return 'Content-Type: {}\r\n'.format(ct).encode('utf-8')

    def __bytes__(self):
        """Return a byte-string representing the form data,
        including attached files.
        """
        buffer = io.BytesIO()
        boundary = b'--' + self.boundary + b'\r\n'

        # Add the form fields
        for name, value in self.form_fields:
            buffer.write(boundary)
            buffer.write(self._form_data(name))
            buffer.write(b'\r\n')
            buffer.write(value.encode('utf-8'))
            buffer.write(b'\r\n')

        # Add the files to upload
        for f_name, filename, f_content_type, body in self.files:
            buffer.write(boundary)
            buffer.write(self._attached_file(f_name, filename))
            buffer.write(self._content_type(f_content_type))
            buffer.write(b'\r\n')
            buffer.write(body)
            buffer.write(b'\r\n')

        buffer.write(b'--' + self.boundary + b'--\r\n')
        return buffer.getvalue()


class TimeoutIO(io.BytesIO):
    def __init__(self, file_: Union[BinaryIO, int], timeout: float):
        super().__init__()
        self.file_ = file_
        self.timeout = timeout
        if isinstance(file_, int):
            self._write = lambda *args, **kwargs: os.write(file_, *args, **kwargs)
            self._read = lambda *args, **kwargs: os.read(file_, *args, **kwargs)
            self._close = lambda: os.close(file_)
            self._flush = lambda: None
        elif isinstance(file_, io.BufferedIOBase):
            self._write = file_.write
            self._read = file_.read
            self._close = file_.close
            self._flush = file_.flush
        else:
            raise TypeError('file_ must be an int or BinaryIO: %', type(file_))

    def read(self, *args, **kwargs) -> bytes:
        rlist, _, _ = select([self.file_], [], [], self.timeout)
        if rlist:
            return self._read(*args, **kwargs)
        else:
            raise TimeoutError('Read timeout ({}s)'.format(self.timeout))

    def write(self, *args, **kwargs) -> int:
        _, wlist, _ = select([], [self.file_], [], self.timeout)
        if wlist:
            return self._write(*args, **kwargs)
        else:
            raise TimeoutError('Write timeout ({}s)'.format(self.timeout))

    def flush(self) -> None:
        self._flush()

    def close(self) -> None:
        self._close()


# http://code.activestate.com/recipes/576655-wait-for-network-service-to-appear/
def wait_net_service(server, port, timeout=None):
    """ Wait for network service to appear
        @param timeout: in seconds, if None or 0 wait forever
        @return: True of False, if timeout is None may return only True or
                 throw unhandled network exception
    """
    s = socket.socket()
    if timeout:
        from time import time as now
        # time module is needed to calc timeout shared between two exceptions
        end = now() + timeout

    while True:
        try:
            if timeout:
                next_timeout = end - now()
                if next_timeout < 0:
                    raise TimeoutError('Timed out: %s' % timeout)
                else:
                    s.settimeout(next_timeout)

            s.connect((server, port))

        except socket.timeout as err:
            # this exception occurs only if timeout is set
            if timeout:
                raise TimeoutError('Timed out: %s' % timeout)

        except socket.error as err:
            # catch timeout exception from underlying network library
            # this one is different from socket.timeout
            if type(err.args) != tuple:  # or err[0] != errno.ETIMEDOUT:
                raise
        else:
            s.close()
            return



from unittest import TestCase

from brotab.operations import apply_move_commands
from brotab.operations import apply_update_commands
from brotab.operations import infer_all_commands
from brotab.operations import infer_move_commands
from brotab.tab import parse_tab_lines


# class TestLIS(TestCase):
#     def test_one(self):
#         seq = [1, 3, 0, 5, 6, 4, 7, 2, 0]
#         result = get_longest_increasing_subsequence(seq)
#         self.assertEqual([1, 3, 5, 6, 7], result)


class TestReconstruct(TestCase):
    def test_move_one_from_start_to_end(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
            'f.0.0\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        self.assertEqual(commands, [(0, 0, 4)])
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_move_one_from_end_to_start(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.4\ta\turl',
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        self.assertEqual(commands, [(4, 0, 0)])
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_move_one_from_start_to_center(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
            'f.0.5\ta\turl',
            'f.0.6\ta\turl',
            'f.0.7\ta\turl',
            'f.0.8\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
            'f.0.0\ta\turl',
            'f.0.5\ta\turl',
            'f.0.6\ta\turl',
            'f.0.7\ta\turl',
            'f.0.8\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        self.assertEqual(commands, [(0, 0, 4)])
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_crossings(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
            'f.0.5\ta\turl',
            'f.0.6\ta\turl',
            'f.0.7\ta\turl',
            'f.0.8\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.4\ta\turl',
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.6\ta\turl',
            'f.0.7\ta\turl',
            'f.0.8\ta\turl',
            'f.0.5\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        self.assertEqual(commands, [(4, 0, 0), (5, 0, 8)])
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_decreasing_ids_from_start_to_end(self):
        before = parse_tab_lines([
            'f.0.10\ta\turl',
            'f.0.9\ta\turl',
            'f.0.8\ta\turl',
            'f.0.7\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.9\ta\turl',
            'f.0.8\ta\turl',
            'f.0.7\ta\turl',
            'f.0.10\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        self.assertEqual(commands, [(10, 0, 3)])
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_move_pair_from_start_to_end(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        self.assertEqual(commands, [(1, 0, 4), (0, 0, 3)])
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_move_pair_from_end_to_start(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        # self.assertEqual(commands, [(1, 0, 4), (0, 0, 3)])
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_move_several_upwards(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
            'f.0.5\ta\turl',
            'f.0.6\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.1\ta\turl',
            'f.0.3\ta\turl',
            'f.0.5\ta\turl',
            'f.0.6\ta\turl',
            'f.0.0\ta\turl',
            'f.0.2\ta\turl',
            'f.0.4\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        # self.assertEqual(commands, [(1, 0, 4), (0, 0, 3)])
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_move_several_downwards(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.2\ta\turl',
            'f.0.3\ta\turl',
            'f.0.4\ta\turl',
            'f.0.5\ta\turl',
            'f.0.6\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.2\ta\turl',
            'f.0.4\ta\turl',
            'f.0.6\ta\turl',
            'f.0.0\ta\turl',
            'f.0.1\ta\turl',
            'f.0.3\ta\turl',
            'f.0.5\ta\turl',
        ])
        commands = infer_move_commands(before, after)
        actual_after = apply_move_commands(before, commands)
        self.assertEqual(actual_after, after)

    def test_move_one_to_another_existing_window_same_index(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
            'f.1.1\ta\turl',
        ])
        after = parse_tab_lines([
            'f.1.0\ta\turl',
            'f.1.1\ta\turl',
        ])
        delete_commands, move_commands, update_commands = infer_all_commands(before, after)
        self.assertGreater(len(move_commands), 0)
        actual_after = apply_move_commands(before, move_commands)
        self.assertEqual(actual_after, after)

    def test_move_one_to_another_existing_window_below(self):
        before = parse_tab_lines([
            'f.0.0\ta1\turl1',
            'f.1.1\ta2\turl2',
            'f.1.2\ta3\turl3',
        ])
        after = parse_tab_lines([
            'f.1.1\ta2\turl2',
            'f.1.2\ta3\turl3',
            'f.1.0\ta1\turl1',
        ])
        delete_commands, move_commands, update_commands = infer_all_commands(before, after)
        self.assertGreater(len(move_commands), 0)
        actual_after = apply_move_commands(before, move_commands)
        self.assertEqual(actual_after, after)

    def test_move_one_to_another_existing_window_above(self):
        before = parse_tab_lines([
            'f.0.0\ta1\turl1',
            'f.1.1\ta2\turl2',
            'f.1.2\ta3\turl3',
        ])
        after = parse_tab_lines([
            'f.0.2\ta3\turl3',
            'f.0.0\ta1\turl1',
            'f.1.1\ta2\turl2',
        ])
        delete_commands, move_commands, update_commands = infer_all_commands(before, after)
        self.assertGreater(len(move_commands), 0)
        actual_after = apply_move_commands(before, move_commands)
        self.assertEqual(actual_after, after)

    def test_move_one_to_another_existing_window_above_2(self):
        before = parse_tab_lines([
            'f.0.0\ta1\turl1',
            'f.0.1\ta2\turl2',
            'f.1.2\ta3\turl3',
            'f.1.3\ta4\turl4',
        ])
        after = parse_tab_lines([
            'f.0.0\ta1\turl1',
            'f.0.1\ta2\turl2',
            'f.0.3\ta4\turl4',
            'f.1.2\ta3\turl3',
        ])
        delete_commands, move_commands, update_commands = infer_all_commands(before, after)
        self.assertGreater(len(move_commands), 0)
        actual_after = apply_move_commands(before, move_commands)
        self.assertEqual(actual_after, after)

    def test_move_one_to_another_existing_window_several_mix(self):
        before = parse_tab_lines([
            'f.0.0\ta1\turl1',
            'f.0.1\ta2\turl2',
            'f.0.2\ta3\turl3',
            'f.1.3\ta3\turl4',
            'f.1.4\ta3\turl5',
            'f.1.5\ta3\turl6',
        ])
        after = parse_tab_lines([
            'f.0.5\ta3\turl6',
            'f.0.2\ta3\turl3',
            'f.1.3\ta3\turl4',
            'f.1.4\ta3\turl5',
            'f.1.0\ta1\turl1',
            'f.1.1\ta2\turl2',
        ])
        delete_commands, move_commands, update_commands = infer_all_commands(before, after)
        self.assertGreater(len(move_commands), 0)
        actual_after = apply_move_commands(before, move_commands)
        self.assertEqual(actual_after, after)

    # def test_move_one_to_another_existing_window_fixture_test1(self):
    #     before = parse_tab_lines(slurp_lines('tests/fixtures/move_to_another_window_test1_before.txt'))
    #     after = parse_tab_lines(slurp_lines('tests/fixtures/move_to_another_window_test1_after.txt'))
    #     delete_commands, move_commands = infer_delete_and_move_commands(before, after)
    #     print('COMMANDS', move_commands)
    #     self.assertGreater(len(move_commands), 0)
    #     actual_after = apply_move_commands(before, move_commands)
    #     self.assertEqual(actual_after, after)


# class TestSequence(TestCase):
#     def test_get_longest_contiguous_increasing_sequence(self):
#         tabs = [
#             'f.0\tFirst',
#             'f.1\tSecond',
#             'f.2\tThird',
#         ]
#         result = get_longest_contiguous_increasing_sequence(tabs)
#         self.assertEqual(result, (0, 3))

#         tabs = [
#             'f.4\te',
#             'f.1\tb',
#             'f.2\tc',
#             'f.3\td',
#             'f.0\ta',
#         ]
#         result = get_longest_contiguous_increasing_sequence(tabs)
#         self.assertEqual(result, (1, 3))

#         tabs = [
#             'f.4\te',
#             'f.1\tb',
#             'f.2\tc',
#             'f.3\td',
#             'f.0\ta',
#             'f.9\ta',
#             'f.5\tz',
#             'f.6\tz',
#             'f.7\tz',
#             'f.8\tz',
#         ]
#         result = get_longest_contiguous_increasing_sequence(tabs)
#         self.assertEqual(result, (6, 4))

#         tabs = [
#             'f.9\te',
#             'f.4\tz',
#             'f.5\tz',
#             'f.6\tz',
#             'f.7\tz',
#             'f.1\tb',
#             'f.2\tc',
#             'f.3\td',
#             'f.0\ta',
#         ]
#         result = get_longest_contiguous_increasing_sequence(tabs)
#         self.assertEqual(result, (1, 4))


class TestInferDeleteMoveUpdateCommands(TestCase):
    def _eq(self, tabs_before, tabs_after, expected_deletes, expected_moves, expected_updates):
        deletes, moves, updates = infer_all_commands(
            parse_tab_lines(tabs_before),
            parse_tab_lines(tabs_after))
        self.assertEqual(expected_deletes, deletes)
        self.assertEqual(expected_moves, moves)
        self.assertEqual(expected_updates, updates)

    def test_only_deletes(self):
        self._eq(
            ['f.0.0\ttitle\turl',
             'f.0.1\ttitle\turl',
             'f.0.2\ttitle\turl'],
            ['f.0.2\ttitle\turl'],
            ['f.0.1', 'f.0.0'],
            [],
            []
        )

        self._eq(
            ['f.0.0\ttitle\turl',
             'f.0.1\ttitle\turl',
             'f.0.2\ttitle\turl',
             'f.0.3\ttitle\turl'],
            ['f.0.0\ttitle\turl',
             'f.0.2\ttitle\turl'],
            ['f.0.3', 'f.0.1'],
            [],
            []
        )

    def test_only_moves(self):
        self._eq(
            [
                'f.0.0\ttitle\turl',
                'f.0.1\ttitle\turl',
                'f.0.2\ttitle\turl',
            ],
            [
                'f.0.2\ttitle\turl',
                'f.0.0\ttitle\turl',
                'f.0.1\ttitle\turl',
            ],
            [],
            [(2, 0, 0)],
            []
        )

        self._eq(
            [
                'f.0.0\ttitle\turl',
                'f.0.1\ttitle\turl',
                'f.0.2\ttitle\turl',
            ],
            [
                'f.0.2\ttitle\turl',
                'f.0.1\ttitle\turl',
                'f.0.0\ttitle\turl',
            ],
            [],
            [(2, 0, 0), (1, 0, 1)],
            []
        )


class TestUpdate(TestCase):
    def test_update_only_url(self):
        before = parse_tab_lines([
            'f.0.0\ta\turl',
        ])
        after = parse_tab_lines([
            'f.0.0\ta\turl_changed',
        ])
        deletes, moves, updates = infer_all_commands(before, after)
        self.assertEqual([], deletes)
        self.assertEqual([], moves)

        actual_after = apply_update_commands(before, updates)
        self.assertEqual(actual_after, after)



"""
This module contains helpers to work with an index of text from a browser.
"""
import argparse
import csv
import ctypes
import logging
import sqlite3
import sys
from contextlib import suppress

MAX_FIELD_LEN = 131072


logger = logging.getLogger('brotab')


def index(sqlite_filename, tsv_filename):
    logger.info('Reading tsv file %s', tsv_filename)
    # https://stackoverflow.com/questions/15063936/csv-error-field-larger-than-field-limit-131072
    # https://github.com/balta2ar/brotab/issues/25
    # It should work on Python 3 and Python 2, on any CPU / OS.
    csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))

    with open(tsv_filename, encoding='utf-8') as tsv_file:
        lines = [tuple(line) for line in csv.reader(tsv_file, delimiter='\t',
                                                    quoting=csv.QUOTE_NONE)]

    logger.info(
        'Creating sqlite DB filename %s from tsv %s (%s lines)',
        sqlite_filename, tsv_filename, len(lines))
    conn = sqlite3.connect(sqlite_filename)
    cursor = conn.cursor()
    with suppress(sqlite3.OperationalError):
        cursor.execute('drop table tabs;')
    cursor.execute(
        'create virtual table tabs using fts5('
        '    tab_id, title, url, body, tokenize="porter unicode61");')
    cursor.executemany('insert into tabs values (?, ?, ?, ?)', lines)
    conn.commit()
    conn.close()


def main():
    parser = argparse.ArgumentParser(description='Index text from tabs')
    parser.add_argument('sqlite', help='output sqlite DB file name')
    parser.add_argument('tsv', help='input tsv file name')
    args = parser.parse_args()

    index(args.sqlite, args.tsv)


if __name__ == '__main__':
    main()



import os
import sqlite3

import psutil


def assert_file_absent(filename):
    if os.path.isfile(filename):
        os.remove(filename)
    assert not os.path.isfile(filename)


def assert_file_not_empty(filename):
    assert os.path.isfile(filename)
    assert os.path.getsize(filename) > 0


def assert_file_contents(filename, expected_contents):
    with open(filename) as file_:
        actual_contents = file_.read()
        assert expected_contents == actual_contents, \
            '"%s" != "%s"' % (expected_contents, actual_contents)


def assert_sqlite3_table_contents(db_filename, table_name, expected_contents):
    conn = sqlite3.connect(db_filename)
    cursor = conn.cursor()
    cursor.execute('select * from %s' % (table_name,))
    actual_contents = '\n'.join(['\t'.join(line) for line in cursor.fetchall()])
    assert expected_contents == actual_contents, \
        '"%s" != "%s"' % (expected_contents, actual_contents)


def kill_by_substring(substring):
    mypid = os.getpid()
    for proc in psutil.process_iter():
        line = ' '.join(proc.cmdline())
        if substring in line and proc.pid != mypid:
            print('>', proc.name(), line)
            proc.kill()


class TimeoutException(Exception):
    pass



from string import ascii_letters
from time import sleep
from typing import List
from unittest import TestCase
from unittest.mock import patch
from uuid import uuid4

from brotab.api import SingleMediatorAPI
from brotab.inout import MIN_MEDIATOR_PORT
from brotab.inout import get_available_tcp_port
from brotab.inout import in_temp_dir
from brotab.inout import spit
from brotab.main import create_clients
from brotab.main import run_commands
from brotab.mediator.const import DEFAULT_HTTP_IFACE
from brotab.mediator.http_server import MediatorHttpServer
from brotab.mediator.remote_api import default_remote_api
from brotab.mediator.transport import Transport
from brotab.tests.utils import assert_file_absent
from brotab.tests.utils import assert_file_contents
from brotab.tests.utils import assert_file_not_empty
from brotab.tests.utils import assert_sqlite3_table_contents


class MockedLoggingTransport(Transport):
    def __init__(self):
        self._sent = []
        self._received = []

    def reset(self):
        self._sent = []
        self._received = []

    @property
    def sent(self):
        return self._sent

    @property
    def received(self):
        return self._received

    def received_extend(self, values) -> None:
        for value in values:
            self._received.append(value)

    def send(self, message) -> None:
        self._sent.append(message)

    def recv(self):
        if self._received:
            return self._received.pop(0)

    def close(self):
        pass


class MockedMediator:
    def __init__(self, prefix='a', port=None, remote_api=None):
        self.port = get_available_tcp_port() if port is None else port
        self.transport = MockedLoggingTransport()
        self.remote_api = default_remote_api(self.transport) if remote_api is None else remote_api
        self.server = MediatorHttpServer(DEFAULT_HTTP_IFACE, self.port, self.remote_api, 0.050)
        self.thread = self.server.run.in_thread()
        self.transport.received_extend(['mocked'])
        self.api = SingleMediatorAPI(prefix, port=self.port, startup_timeout=1)
        assert self.api.browser == 'mocked'
        self.transport.reset()

    def join(self):
        self.server.shutdown()
        self.thread.join()

    def __enter__(self):
        return self

    def __exit__(self, type_, value, tb):
        self.join()


class DummyBrowserRemoteAPI:
    """
    Dummy version of browser API for integration smoke tests.
    """

    def list_tabs(self):
        return ['1.1\ttitle\turl']

    def query_tabs(self, query_info: str):
        raise NotImplementedError()

    def move_tabs(self, move_triplets: str):
        raise NotImplementedError()

    def open_urls(self, urls: List[str], window_id=None):
        raise NotImplementedError()

    def close_tabs(self, tab_ids: str):
        raise NotImplementedError()

    def new_tab(self, query):
        raise NotImplementedError()

    def activate_tab(self, tab_id: int, focused: bool):
        raise NotImplementedError()

    def get_active_tabs(self) -> str:
        return '1.1'

    def get_words(self, tab_id, match_regex, join_with):
        return ['a', 'b']

    def get_text(self, delimiter_regex, replace_with):
        return ['1.1\ttitle\turl\tbody']

    def get_html(self, delimiter_regex, replace_with):
        return ['1.1\ttitle\turl\t<body>some body</body>']

    def get_browser(self):
        return 'mocked'


def run_mocked_mediators(count, default_port_offset, delay):
    """
    How to run:

    python -c 'from brotab.tests.test_main import run_mocked_mediators as run; run(3, 0, 0)'
    python -c 'from brotab.tests.test_main import run_mocked_mediators as run; run(count=3, default_port_offset=10, delay=0)'
    """
    assert count > 0
    print('Creating %d mediators' % count)
    start_port = MIN_MEDIATOR_PORT + default_port_offset
    ports = range(start_port, start_port + count)
    mediators = [MockedMediator(letter, port, DummyBrowserRemoteAPI())
                 for i, letter, port in zip(range(count), ascii_letters, ports)]
    sleep(delay)
    print('Ready')
    for mediator in mediators:
        print(mediator.port)
    mediators[0].thread.join()


def run_mocked_mediator_current_thread(port):
    """
    How to run:

    python -c 'from brotab.tests.test_main import run_mocked_mediator_current_thread as run; run(4635)'
    """
    remote_api = DummyBrowserRemoteAPI()
    port = get_available_tcp_port() if port is None else port
    server = MediatorHttpServer(DEFAULT_HTTP_IFACE, port, remote_api, 0.050)
    server.run.here()


class WithMediator(TestCase):
    def setUp(self):
        self.mediator = MockedMediator('a')

    def tearDown(self):
        self.mediator.join()

    def _run_commands(self, commands):
        with patch('brotab.main.get_mediator_ports') as mocked:
            mocked.side_effect = [range(self.mediator.port, self.mediator.port + 1)]
            return run_commands(commands)

    def _assert_init(self):
        """Pop get_browser commands from the beginning until we have none."""
        expected = {'name': 'get_browser'}
        popped = 0
        while self.mediator.transport.sent:
            if expected != self.mediator.transport.sent[0]:
                break
            self.mediator.transport.sent.pop(0)
            popped += 1
        assert popped > 0, 'Expected to pop at least one get_browser command'


class TestCreateClients(WithMediator):
    def test_default_target_hosts(self):
        with patch('brotab.main.get_mediator_ports') as mocked:
            mocked.side_effect = [range(self.mediator.port, self.mediator.port + 1)]
            clients = create_clients()
        assert 1 == len(clients)
        assert self.mediator.port == clients[0]._port

    def test_one_custom_target_hosts(self):
        clients = create_clients('127.0.0.1:%d' % self.mediator.port)
        assert 1 == len(clients)
        assert self.mediator.port == clients[0]._port

    def test_two_custom_target_hosts(self):
        clients = create_clients('127.0.0.1:%d,localhost:%d' %
                                 (self.mediator.port, self.mediator.port))
        assert 2 == len(clients)
        assert self.mediator.port == clients[0]._port
        assert self.mediator.port == clients[1]._port


class TestActivate(WithMediator):
    def test_activate_ok(self):
        self._run_commands(['activate', 'a.1.2'])
        self._assert_init()
        assert self.mediator.transport.sent == [
            {'name': 'activate_tab', 'tab_id': 2, 'focused': False}
        ]

    def test_activate_focused_ok(self):
        self._run_commands(['activate', '--focused', 'a.1.2'])
        self._assert_init()
        assert self.mediator.transport.sent == [
            {'name': 'activate_tab', 'tab_id': 2, 'focused': True}
        ]


class TestText(WithMediator):
    def test_text_no_arguments_ok(self):
        self.mediator.transport.received_extend([
            'mocked',
            ['1.1\ttitle\turl\tbody'],
        ])

        output = []
        with patch('brotab.main.stdout_buffer_write', output.append):
            self._run_commands(['text'])
        self._assert_init()
        assert self.mediator.transport.sent == [
            {'delimiter_regex': '/\\n|\\r|\\t/g', 'name': 'get_text', 'replace_with': '" "'},
        ]
        assert output == [b'a.1.1\ttitle\turl\tbody\n']

    def test_text_with_tab_id_ok(self):
        self.mediator.transport.received_extend([
            'mocked',
            [
                '1.1\ttitle\turl\tbody',
                '1.2\ttitle\turl\tbody',
                '1.3\ttitle\turl\tbody',
            ],
        ])

        output = []
        with patch('brotab.main.stdout_buffer_write', output.append):
            self._run_commands(['text', 'a.1.2', 'a.1.3'])
        self._assert_init()
        assert self.mediator.transport.sent == [
            {'delimiter_regex': '/\\n|\\r|\\t/g', 'name': 'get_text', 'replace_with': '" "'},
        ]
        assert output == [b'a.1.2\ttitle\turl\tbody\na.1.3\ttitle\turl\tbody\n']


class TestHtml(WithMediator):
    def test_html_no_arguments_ok(self):
        self.mediator.transport.received_extend([
            'mocked',
            ['1.1\ttitle\turl\tbody'],
        ])

        output = []
        with patch('brotab.main.stdout_buffer_write', output.append):
            self._run_commands(['html'])
        self._assert_init()
        assert self.mediator.transport.sent == [
            {'delimiter_regex': '/\\n|\\r|\\t/g', 'name': 'get_html', 'replace_with': '" "'},
        ]
        assert output == [b'a.1.1\ttitle\turl\tbody\n']

    def test_html_with_tab_id_ok(self):
        self.mediator.transport.received_extend([
            'mocked',
            [
                '1.1\ttitle\turl\tbody',
                '1.2\ttitle\turl\tbody',
                '1.3\ttitle\turl\tbody',
            ],
        ])

        output = []
        with patch('brotab.main.stdout_buffer_write', output.append):
            self._run_commands(['html', 'a.1.2', 'a.1.3'])
        self._assert_init()
        assert self.mediator.transport.sent == [
            {'delimiter_regex': '/\\n|\\r|\\t/g', 'name': 'get_html', 'replace_with': '" "'},
        ]
        assert output == [b'a.1.2\ttitle\turl\tbody\na.1.3\ttitle\turl\tbody\n']


class TestIndex(WithMediator):
    def test_index_no_arguments_ok(self):
        self.mediator.transport.received_extend([
            'mocked',
            ['1.1\ttitle\turl\tbody'],
        ])

        sqlite_filename = in_temp_dir('tabs.sqlite')
        tsv_filename = in_temp_dir('tabs.tsv')
        assert_file_absent(sqlite_filename)
        assert_file_absent(tsv_filename)
        output = []
        with patch('brotab.main.stdout_buffer_write', output.append):
            self._run_commands(['index'])
        self._assert_init()
        assert self.mediator.transport.sent == [
            {'delimiter_regex': '/\\n|\\r|\\t/g',
             'name': 'get_text', 'replace_with': '" "'},
        ]
        assert_file_not_empty(sqlite_filename)
        assert_file_not_empty(tsv_filename)
        assert_file_contents(tsv_filename, 'a.1.1\ttitle\turl\tbody\n')
        assert_sqlite3_table_contents(
            sqlite_filename, 'tabs', 'a.1.1\ttitle\turl\tbody')

    def test_index_custom_filename(self):
        self.mediator.transport.received_extend([
            'mocked',
            ['1.1\ttitle\turl\tbody'],
        ])

        sqlite_filename = in_temp_dir(uuid4().hex + '.sqlite')
        tsv_filename = in_temp_dir(uuid4().hex + '.tsv')
        assert_file_absent(sqlite_filename)
        assert_file_absent(tsv_filename)
        spit(tsv_filename, 'a.1.1\ttitle\turl\tbody\n')

        output = []
        with patch('brotab.main.stdout_buffer_write', output.append):
            self._run_commands(
                ['index', '--sqlite', sqlite_filename, '--tsv', tsv_filename])
        assert self.mediator.transport.sent == []
        assert_file_not_empty(sqlite_filename)
        assert_file_not_empty(tsv_filename)
        assert_file_contents(tsv_filename, 'a.1.1\ttitle\turl\tbody\n')
        assert_sqlite3_table_contents(
            sqlite_filename, 'tabs', 'a.1.1\ttitle\turl\tbody')
        assert_file_absent(sqlite_filename)
        assert_file_absent(tsv_filename)


class TestOpen(WithMediator):
    def test_three_urls_ok(self):
        self.mediator.transport.received_extend([
            'mocked',
            ['1.1', '1.2', '1.3'],
        ])

        urls = ['url1', 'url2', 'url3']
        output = []
        with patch('brotab.main.stdout_buffer_write', output.append):
            with patch('brotab.main.read_stdin_lines', return_value=urls):
                self._run_commands(['open', 'a.1'])
        self._assert_init()
        assert self.mediator.transport.sent == [
            {'name': 'open_urls', 'urls': ['url1', 'url2', 'url3'], 'window_id': 1},
        ]
        assert output == [b'a.1.1\na.1.2\na.1.3\n']



import re
import shutil
from base64 import urlsafe_b64decode
from base64 import urlsafe_b64encode
from os.path import expanduser
from os.path import expandvars
from os.path import getsize


def split_tab_ids(string):
    items = re.split(r'[ \t\r\n]+', string)
    return list(filter(None, items))


def encode_query(string):
    return str(urlsafe_b64encode(string.encode('utf-8')), 'utf-8')


def decode_query(string):
    return urlsafe_b64decode(string).decode('utf-8')


def get_file_size(path):
    try:
        return getsize(path)
    except FileNotFoundError:
        return None


def which(program):
    paths = [None, '/usr/local/bin', '/usr/bin', '/bin', '~/bin', '~/.local/bin']
    for path in paths:
        path = expanduser(expandvars(path)) if path else None
        path = shutil.which(program, path=path)
        if path:
            return path



import os
import signal
import threading
from http.server import BaseHTTPRequestHandler
from http.server import HTTPServer
from json import dumps
from subprocess import Popen
from subprocess import check_output
from unittest import TestCase
from urllib.parse import parse_qs
from urllib.parse import urlparse

import pytest

from brotab.api import api_must_ready
from brotab.inout import get_available_tcp_port
from brotab.inout import wait_net_service
from brotab.mediator.const import DEFAULT_MIN_HTTP_PORT
from brotab.operations import make_update
from brotab.tab import parse_tab_lines


def run(args):
    return check_output(args, shell=True).decode('utf-8').strip().splitlines()


def git_root():
    return run(['git rev-parse --show-toplevel'])[0]


def requires_integration_env():
    value = os.environ.get('INTEGRATION_TEST')
    return pytest.mark.skipif(
        value is None,
        reason=f"Skipped because INTEGRATION_TEST=1 is not set"
    )


TIMEOUT = 60  # 15
ECHO_SERVER_PORT = 8087


class EchoRequestHandler(BaseHTTPRequestHandler):
    """
    Sample URL:

        localhost:9000?title=tab1&body=tab1
    """

    def _get_str_arg(self, path, arg_name):
        args = parse_qs(urlparse(path).query)
        return ''.join(args.get(arg_name, ''))

    def do_GET(self):
        title = self._get_str_arg(self.path, 'title')
        body = self._get_str_arg(self.path, 'body')
        print('EchoServer received TITLE "%s" BODY "%s"' % (title, body))

        self.send_response(200)
        self.send_header("Content-Type", "text/html; charset=utf-8")
        reply = ('<html><head><title>%s</title></head>'
                 '<body>%s</body></html>'
                 % (title, body)).encode('utf-8')
        self.send_header("Content-Length", str(len(reply)))
        self.end_headers()
        self.wfile.write(reply)
        # self.wfile.close()


ECHO_SERVER_HOST = 'localhost'
ECHO_SERVER_PORT = 9000


class EchoServer:
    """
    This EchoServer is used to customize page title and content using URL
    parameters.
    """

    def __init__(self):
        self._thread = None
        self._server = None

    def run(self, host=ECHO_SERVER_HOST, port=ECHO_SERVER_PORT):
        self._server = HTTPServer((host, port), EchoRequestHandler)
        self._thread = threading.Thread(
            target=self._server.serve_forever, daemon=True)
        self._thread.start()

    def stop(self):
        self._server.shutdown()
        self._server.socket.close()
        self._thread.join(TIMEOUT)

    @staticmethod
    def url(title='', body=''):
        return 'http://%s:%s?title=%s&body=%s' % (
            ECHO_SERVER_HOST, ECHO_SERVER_PORT, title, body)


class Brotab:
    def __init__(self, target_hosts: str):
        """
        target_hosts: e.g. 'localhost:4625,localhost:4626'
        """
        self.targets = target_hosts
        self.options = '--target %s' % self.targets if self.targets else ''

    def list(self):
        return run(f'bt {self.options} list')

    def tabs(self):
        return parse_tab_lines(self.list())

    def open(self, window_id, url):
        return run(f'echo "{url}" | bt {self.options} open {window_id}')

    def active(self):
        return run(f'bt {self.options} active')

    def windows(self):
        return run(f'bt {self.options} windows')

    def navigate(self, tab_id, url):
        return run(f'bt {self.options} navigate {tab_id} "{url}"')

    def update_stdin(self, updates):
        updates = dumps(updates)
        return run(f'echo \'{updates}\' | bt {self.options} update')

    def update(self, args):
        return run(f'bt {self.options} update {args}')


class Browser:
    CMD = ''
    CWD = ''
    PROFILE = ''

    def __init__(self):
        print('CMD', self.CMD, 'CWD', self.CWD)
        # Used a trick from here: https://stackoverflow.com/a/22582602/258421
        os.makedirs('/dev/shm/%s' % self.PROFILE, exist_ok=True)
        self._browser = Popen(self.CMD, shell=True,
                              cwd=self.CWD, preexec_fn=os.setsid)
        print('PID', self._browser.pid)
        wait_net_service('localhost', DEFAULT_MIN_HTTP_PORT, TIMEOUT)
        print('init done PID', self._browser.pid)

    def stop(self):
        os.killpg(os.getpgid(self._browser.pid), signal.SIGTERM)
        self._browser.wait(TIMEOUT)

    @property
    def pid(self):
        return self._browser.pid


class Container:
    NAME = 'chrome/chromium'

    def __init__(self):
        root = git_root()
        self.guest_port = 4625
        self.host_port = get_available_tcp_port()
        display = os.environ.get('DISPLAY', ':0')
        args = ['docker', 'run', '-v',
                f'"{root}:/brotab"',
                # '-p', '19222:9222',
                '-p', f'{self.host_port}:{self.guest_port}',
                '--detach --rm --cpuset-cpus 0',
                '--memory 512mb -v /tmp/.X11-unix:/tmp/.X11-unix',
                f'-e DISPLAY=unix{display}',
                '-v /dev/shm:/dev/shm',
                'brotab-integration']
        cmd = ' '.join(args)
        self.container_id = run(cmd)[0]
        api_must_ready(self.host_port, self.NAME, 'a', client_timeout=3.0, startup_timeout=10.0)

    def stop(self):
        run(f'docker kill {self.container_id}')

    def __enter__(self):
        return self

    def __exit__(self, type_, value, tb):
        self.stop()

    @property
    def guest_addr(self):
        return f'localhost:{self.guest_port}'

    @property
    def host_addr(self):
        return f'localhost:{self.host_port}'

    def echo_url(self, title=None, body=None):
        url = f'http://{self.guest_addr}/echo?'
        url += 'title=' + title if title else ''
        url += '&body=' + body if body else ''
        return url


def targets(containers: [Container]) -> str:
    return ','.join([c.host_addr for c in containers])


@requires_integration_env()
class TestIntegration(TestCase):
    def test_open_single(self):
        with Container() as c:
            bt = Brotab(targets([c]))
            tabs = bt.list()
            assert 'tab1' not in ''.join(tabs)
            tab_ids = bt.open('a.1', c.echo_url('tab1'))
            assert len(tab_ids) == 1

            tabs = bt.list()
            assert 'tab1' in ''.join(tabs)
            assert tab_ids[0] in ''.join(tabs)

    def test_active_tabs(self):
        with Container() as c:
            bt = Brotab(targets([c]))
            bt.open('a.1', c.echo_url('tab1'))
            bt.open('a.1', c.echo_url('tab2'))
            bt.open('a.1', c.echo_url('tab3'))
            assert len(bt.tabs()) == 4
            active_id = bt.active()[0].split('\t')[0]
            assert active_id == bt.tabs()[-1].id

    def test_navigate_single(self):
        with Container() as c:
            bt = Brotab(targets([c]))
            tab_ids = bt.open('a.1', c.echo_url('tab1'))
            assert len(tab_ids) == 1

            tabs = bt.list()
            assert 'tab1' in ''.join(tabs)
            assert tab_ids[0] in ''.join(tabs)

            bt.navigate(tab_ids[0], c.echo_url('tab2'))
            tabs = bt.list()
            assert 'tab2' in ''.join(tabs)
            assert tab_ids[0] in ''.join(tabs)

    def test_update_three(self):
        with Container() as c:
            bt = Brotab(targets([c]))
            bt.open('a.1', c.echo_url('tab1'))
            bt.open('a.1', c.echo_url('tab1'))
            lines = sorted(bt.list())
            assert len(lines) == 3
            assert 'tab2' not in ''.join(lines)

            tabs = parse_tab_lines(lines)
            bt.update_stdin([make_update(tabId=tabs[0].id, url=c.echo_url('tab2'))])
            bt.update_stdin([make_update(tabId=tabs[1].id, url=c.echo_url('tab2'))])
            bt.update('-tabId {0} -url="{1}"'.format(tabs[2].id, c.echo_url('tab2')))

            lines = bt.list()
            assert 'tab1' not in ''.join(lines)
            assert 'tab2' in ''.join(lines)


@pytest.mark.skip
class TestChromium(TestCase):
    def setUp(self):
        self._echo_server = EchoServer()
        self._echo_server.run()
        self.addCleanup(self._echo_server.stop)

        self._browser = Chromium()
        # self._browser = Firefox()
        # self.addCleanup(self._browser.stop)
        print('SETUP DONE:', self._browser.pid)

    def tearDown(self):
        print('CHROME', self._browser)
        print('BLOCK DONE')

    def test_open_single(self):
        print('SINGLE START')

        tabs = Brotab.list()
        assert 'tab1' not in ''.join(tabs)
        Brotab.open('a.1', EchoServer.url('tab1'))

        tabs = Brotab.list()
        assert 'tab1' in ''.join(tabs)

        print('SINGLE END')

    def test_active_tabs(self):
        Brotab.open('a.1', EchoServer.url('tab1'))
        Brotab.open('a.2', EchoServer.url('tab2'))
        Brotab.open('a.3', EchoServer.url('tab3'))
        assert len(Brotab.tabs()) == 4
        assert Brotab.active()[0] == Brotab.tabs()[-1].id


if __name__ == '__main__':
    server = EchoServer()
    server.run(ECHO_SERVER_HOST, ECHO_SERVER_PORT)
    print('Running EchoServer at %s:%s. Press Enter to terminate' % (
        ECHO_SERVER_HOST, ECHO_SERVER_PORT))
    input()



import os
import logging
import platform


logger = logging.getLogger('brotab')


def make_windows_path(path):
    return path.replace('/', os.sep)


def make_windows_path_double_sep(path):
    return path.replace('/', os.sep * 2).replace('\\', os.sep * 2)


def windows_registry_set_key(key_path, value):
    from winreg import CreateKey, SetValue, HKEY_CURRENT_USER, REG_SZ
    with CreateKey(HKEY_CURRENT_USER, key_path) as sub_key:
        SetValue(sub_key, None, REG_SZ, value)


def register_native_manifest_windows_chrome(manifest_filename):
    key_path = r'Software\Google\Chrome\NativeMessagingHosts\brotab_mediator'
    manifest_filename = make_windows_path(manifest_filename)
    logger.info('Setting registry key "%s" to "%s"', key_path, manifest_filename)
    print('Setting registry key "%s" to "%s"' % (key_path, manifest_filename))
    windows_registry_set_key(key_path, manifest_filename)

def register_native_manifest_windows_brave(manifest_filename):
    key_path = r'Software\BraveSoftware\Brave-Browser\NativeMessagingHosts\brotab_mediator'
    manifest_filename = make_windows_path(manifest_filename)
    logger.info('Setting registry key "%s" to "%s"', key_path, manifest_filename)
    print('Setting registry key "%s" to "%s"' % (key_path, manifest_filename))
    windows_registry_set_key(key_path, manifest_filename)


def register_native_manifest_windows_firefox(manifest_filename):
    key_path = r'Software\Mozilla\NativeMessagingHosts\brotab_mediator'
    manifest_filename = make_windows_path(manifest_filename)
    logger.info('Setting registry key "%s" to "%s"', key_path, manifest_filename)
    print('Setting registry key "%s" to "%s"' % (key_path, manifest_filename))
    windows_registry_set_key(key_path, manifest_filename)


def is_windows() -> bool:
    return platform.system() == 'Windows'


def get_editor() -> str:
    mapping = {
        'Windows': 'notepad'
    }
    system = platform.system()
    editor = mapping.get(system, 'nvim')
    return os.environ.get('EDITOR', editor)



import time


class Waiter:
    def __init__(self, condition):
        self.condition = condition

    def wait(self, timeout: float) -> bool:
        expires_at = time.time() + timeout
        while time.time() < expires_at:
            if self.condition():
                return True
            time.sleep(0.050)
        return False


class ConditionTrue:
    def __init__(self, f):
        self.f = f

    def __call__(self):
        return self.f()


class ConditionRaises:
    def __init__(self, e):
        self.e = e

    def __call__(self, f) -> bool:
        try:
            f()
        except self.e as e:
            return True
        return False



import signal
from typing import Callable

from brotab.mediator.log import mediator_logger


def setup(shutdown: Callable[[], None]) -> None:
    def handler(signum, _frame):
        mediator_logger.info('Got signal %s', signum)
        shutdown()

    signal.signal(signal.SIGINT, handler)
    signal.signal(signal.SIGTERM, handler)



from typing import List
from urllib.parse import quote_plus

from brotab.mediator.log import mediator_logger
from brotab.mediator.transport import Transport


class BrowserRemoteAPI:
    """
    Communicates with a browser using stdin/stdout. This mediator is supposed
    to be run by the browser after a request from the helper extension.
    """

    def __init__(self, transport: Transport):
        self._transport: Transport = transport

    def list_tabs(self):
        command = {'name': 'list_tabs'}
        self._transport.send(command)
        return self._transport.recv()

    def query_tabs(self, query_info: str):
        mediator_logger.info('query info: %s', query_info)
        command = {'name': 'query_tabs', 'query_info': query_info}
        self._transport.send(command)
        return self._transport.recv()

    def move_tabs(self, move_triplets: str):
        """
        :param move_triplets: Comma-separated list of:
            <tabID> <windowID> <newIndex>
        """
        mediator_logger.info('move_tabs, move_triplets: %s', move_triplets)

        triplets = [list(map(int, triplet.split(' ')))
                    for triplet in move_triplets.split(',')]
        mediator_logger.info('moving tab ids: %s', triplets)
        command = {'name': 'move_tabs', 'move_triplets': triplets}
        self._transport.send(command)
        return self._transport.recv()

    def open_urls(self, urls: List[str], window_id=None):
        """
        Open specified list of URLs in a window, specified by window_id.

        If window_id is None, currently active window is used.
        """
        mediator_logger.info('open urls: %s', urls)

        command = {'name': 'open_urls', 'urls': urls}
        if window_id is not None:
            command['window_id'] = window_id
        self._transport.send(command)
        return self._transport.recv()

    def update_tabs(self, updates: [object]):
        """
        Sends a list of updates to the browser. Format:
        [ {
            'tab_id': <tab_id>,
            'properties': {
                'url': <url>,
            }
        } ]
        see https://developer.mozilla.org/en-US/docs/Mozilla/Add-ons/WebExtensions/API/tabs/update
        """
        mediator_logger.info('update tabs: %s', updates)
        command = {'name': 'update_tabs', 'updates': updates}
        self._transport.send(command)
        return self._transport.recv()

    def close_tabs(self, tab_ids: str):
        """
        :param tab_ids: Comma-separated list of tab IDs to close.
        """
        int_tab_ids = [int(id_) for id_ in tab_ids.split(',')]
        mediator_logger.info('closing tab ids: %s', int_tab_ids)
        command = {'name': 'close_tabs', 'tab_ids': int_tab_ids}
        self._transport.send(command)
        return self._transport.recv()

    def new_tab(self, query: str):
        url = "https://www.google.com/search?q=%s" % quote_plus(query)
        mediator_logger.info('opening url: %s', url)
        command = {'name': 'new_tab', 'url': url}
        self._transport.send(command)
        return self._transport.recv()

    def activate_tab(self, tab_id: int, focused: bool):
        mediator_logger.info('activating tab id: %s', tab_id)
        command = {'name': 'activate_tab', 'tab_id': tab_id, 'focused': focused}
        self._transport.send(command)

    def get_active_tabs(self) -> str:
        mediator_logger.info('getting active tabs')
        command = {'name': 'get_active_tabs'}
        self._transport.send(command)
        return self._transport.recv()

    def get_words(self, tab_id: str, match_regex: str, join_with: str):
        mediator_logger.info('getting tab words: %s', tab_id)
        command = {
            'name': 'get_words',
            'tab_id': tab_id,
            'match_regex': match_regex,
            'join_with': join_with,
        }
        self._transport.send(command)
        return self._transport.recv()

    def get_text(self, delimiter_regex: str, replace_with: str):
        mediator_logger.info('getting text, delimiter_regex=%s, replace_with=%s',
                             delimiter_regex, replace_with)
        command = {
            'name': 'get_text',
            'delimiter_regex': delimiter_regex,
            'replace_with': replace_with,
        }
        self._transport.send(command)
        return self._transport.recv()

    def get_html(self, delimiter_regex: str, replace_with: str):
        mediator_logger.info('getting html, delimiter_regex=%s, replace_with=%s',
                             delimiter_regex, replace_with)
        command = {
            'name': 'get_html',
            'delimiter_regex': delimiter_regex,
            'replace_with': replace_with,
        }
        self._transport.send(command)
        return self._transport.recv()

    def get_browser(self):
        mediator_logger.info('getting browser name')
        command = {'name': 'get_browser'}
        self._transport.send(command)
        return self._transport.recv()


def default_remote_api(transport: Transport) -> BrowserRemoteAPI:
    return BrowserRemoteAPI(transport)



import logging
import logging.handlers
from traceback import format_stack

from brotab.inout import in_temp_dir


def _init_logger(tag, filename: str):
    FORMAT = '%(asctime)-15s %(process)-5d %(levelname)-8s %(filename)s:%(lineno)d:%(funcName)s %(message)s'
    MAX_LOG_SIZE = 50 * 1024 * 1024
    LOG_BACKUP_COUNT = 1

    log = logging.getLogger('brotab')
    log.setLevel(logging.DEBUG)
    handler = logging.handlers.RotatingFileHandler(
        filename=filename,
        maxBytes=MAX_LOG_SIZE,
        backupCount=LOG_BACKUP_COUNT,
    )
    handler.setFormatter(logging.Formatter(FORMAT))
    log.addHandler(handler)
    log.info('Logger has been created (%s)', tag)
    return log


def init_brotab_logger(tag: str):
    return _init_logger(tag, in_temp_dir('brotab.log'))


def init_mediator_logger(tag: str):
    return _init_logger(tag, in_temp_dir('brotab_mediator.log'))


def disable_logging():
    # disables flask request logging
    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)
    log.disabled = True
    # TODO: investigate this, maybe we can redirect werkzeug from stdout to a file
    # log.handlers = []
    # disables my own logging in log_and_suppress_exceptions
    # app.logger.disabled = True
    # from flask.logging import default_handler
    # app.logger.removeHandler(default_handler)


def disable_click_echo():
    """Stupid flask started using click which unconditionally prints stupid
    messages"""

    def numb_echo(*args, **kwargs):
        pass

    import click
    click.echo = numb_echo
    click.secho = numb_echo


def stack():
    return '\n'.join(format_stack())


mediator_logger = init_mediator_logger('mediator')
brotab_logger = init_brotab_logger('brotab')



import io
import json
import logging
import socket
import sys
from collections.abc import Mapping
from copy import deepcopy
from functools import partial
from http.client import RemoteDisconnected
from json import dumps
from traceback import print_exc
from typing import List
from urllib.error import HTTPError
from urllib.error import URLError
from urllib.parse import quote_plus
from urllib.request import Request
from urllib.request import urlopen

from brotab.inout import MultiPartForm
from brotab.inout import edit_tabs_in_editor
from brotab.mediator.const import DEFAULT_HTTP_IFACE
from brotab.operations import infer_all_commands
from brotab.parallel import call_parallel
from brotab.tab import parse_tab_lines
from brotab.utils import encode_query
from brotab.wait import ConditionTrue
from brotab.wait import Waiter

logger = logging.getLogger('brotab')

HTTP_TIMEOUT = 10.0
MAX_NUMBER_OF_TABS = 5000


class HttpClient:
    def __init__(self, host='localhost', port=4625, timeout=HTTP_TIMEOUT):
        self._host: str = host
        self._port: int = port
        self._timeout: float = timeout

    def get(self, path, data=None):
        url = 'http://%s:%s%s' % (self._host, self._port, path)
        logger.info('GET %s' % url)
        if data is not None:
            data = data.encode('utf8')
        request = Request(url=url, data=data, method='GET')

        with urlopen(request, timeout=self._timeout) as response:
            return response.read().decode('utf8')

    def post(self, path, files=None):
        url = 'http://%s:%s%s' % (self._host, self._port, path)
        logger.info('POST %s' % url)
        form = MultiPartForm()
        for filename, content in files.items():
            form.add_file(filename, filename,
                          io.BytesIO(content.encode('utf8')))

        data = bytes(form)
        request = Request(url=url, data=data, method='POST')
        request.add_header('Content-Type', form.get_content_type())
        request.add_header('Content-Length', str(len(data)))

        with urlopen(request, timeout=self._timeout) as response:
            return response.read().decode('utf8')


class StartupTimeout(BaseException):
    pass


class SingleMediatorAPI(object):
    """
    This API is designed to work with a single mediator.
    """

    def __init__(self, prefix, host='localhost', port=4625, startup_timeout: float = None, client: HttpClient = None):
        self._prefix = '%s.' % prefix
        self._host = host
        self._port = port
        self._client = HttpClient(host=host, port=port) if client is None else client
        if startup_timeout is not None:
            self.must_ready(timeout=startup_timeout)
        self._pid = self.get_pid()
        self._browser = self.get_browser()

    def must_ready(self, timeout: float) -> None:
        condition = ConditionTrue(lambda: self.get_pid() != -1)
        if not Waiter(condition).wait(timeout=timeout):
            raise StartupTimeout('Failed to start in %s seconds' % timeout)

    def pid_ready(self) -> bool:
        return self.get_pid() != -1

    def pid_not_ready(self) -> bool:
        return not self.pid_ready()

    @property
    def browser(self) -> str:
        return self._browser

    @property
    def ready(self) -> bool:
        return self._browser != '<ERROR>'

    def __str__(self):
        return '%s\t%s:%s\t%s\t%s' % (
            self._prefix, self._host, self._port, self._pid, self._browser)

    def prefix_tab(self, tab):
        return '%s%s' % (self._prefix, tab)

    def prefix_tabs(self, tabs):
        return list(map(self.prefix_tab, tabs))

    def unprefix_tabs(self, tabs):
        num = len(self._prefix)
        return [tab[num:]
                if tab.startswith(self._prefix)
                else tab for tab in tabs]

    def prefix_match(self, tab):
        return tab.startswith(self._prefix)

    def filter_tabs(self, tabs):
        # N = len(self._prefix)
        # return [tab[N:] for tab in tabs
        return [tab for tab in tabs if self.prefix_match(tab)]

    def _split_tabs(self, tabs):
        return [tab.split('.') for tab in tabs]

    def get_pid(self):
        """Get process ID from the mediator."""
        try:
            return int(self._get('/get_pid'))
        except (URLError, HTTPError, socket.timeout, RemoteDisconnected, ConnectionResetError) as e:
            logger.info('_get_pid failed: %s', e)
        return -1

    def get_browser(self):
        """Get browser name from the mediator."""
        try:
            return self._get('/get_browser')
        except (URLError, HTTPError, socket.timeout, RemoteDisconnected, ConnectionResetError) as e:
            logger.info('_get_browser failed: %s', e)
        return '<ERROR>'

    def close_tabs(self, args):
        tabs = ','.join(tab_id for _prefix, _window_id,
                                   tab_id in self._split_tabs(args))
        return self._get('/close_tabs/%s' % tabs)

    def activate_tab(self, args: List[str], focused: bool):
        if len(args) == 0:
            return

        # args: ['a.1.2']
        _prefix, _window_id, tab_id = args[0].split('.')
        self._get('/activate_tab/%s%s' % (tab_id, '?focused=1' if focused else ''))

    def get_active_tabs(self, args) -> List[str]:
        return [self.prefix_tab(tab) for tab in self._get('/get_active_tabs').split(',')]

    def query_tabs(self, args):
        query = args
        if isinstance(query, str):
            try:
                query = json.loads(query)
                if not isinstance(query, Mapping):
                    raise json.JSONDecodeError("json has attributes unsupported by brotab.", "", 0)
            except json.JSONDecodeError as e:
                print("Cannot decode JSON: %s: %s" % (__name__, e), file=sys.stderr)
                return []

        result = self._get('/query_tabs/%s' % encode_query(json.dumps(query)))
        lines = result.splitlines()[:MAX_NUMBER_OF_TABS]
        return self.prefix_tabs(lines)

    def query_tabs_safe(self, args, print_error=False):
        args = args or []
        tabs = []
        try:
            tabs = self.query_tabs(args)
        except ValueError as e:
            print("Cannot decode JSON: %s: %s" % (self, e), file=sys.stderr)
            if print_error:
                print_exc(file=sys.stderr)
        except URLError as e:
            print("Cannot access API %s: %s" % (self, e), file=sys.stderr)
            if print_error:
                print_exc(file=sys.stderr)
        return tabs

    def list_tabs(self, args):
        num_tabs = MAX_NUMBER_OF_TABS
        if len(args) > 0:
            num_tabs = int(args[0])

        result = self._get('/list_tabs')
        lines = []
        for line in result.splitlines()[:num_tabs]:
            lines.append(line)
        return self.prefix_tabs(lines)

    def list_tabs_safe(self, args, print_error=False):
        args = args or []
        tabs = []
        try:
            tabs = self.list_tabs(args)
        except ValueError as e:
            print("Cannot decode JSON: %s: %s" % (self, e), file=sys.stderr)
            if print_error:
                print_exc(file=sys.stderr)
        except URLError as e:
            print("Cannot access API %s: %s" % (self, e), file=sys.stderr)
            if print_error:
                print_exc(file=sys.stderr)
        return tabs

    def move_tabs(self, args):
        logger.info('SENDING MOVE COMMANDS: %s', args)
        commands = ','.join(
            '%s %s %s' % (tab_id, window_id, new_index)
            for tab_id, window_id, new_index in args)
        return self._get('/move_tabs/%s' % quote_plus(commands))

    def open_urls(self, urls, window_id=None):
        data = '\n'.join(urls)
        logger.info('SingleMediatorAPI: open_urls: %s', urls)
        files = {'urls': data}
        ids = self._post('/open_urls'
                         if window_id is None
                         else ('/open_urls/%s' % window_id),
                         files)
        return self.prefix_tabs(ids.splitlines())

    def update_tabs(self, updates):
        logger.info('SingleMediatorAPI: update_tabs: %s', updates)
        files = {'updates': dumps(updates)}
        ids = self._post('/update_tabs', files)
        return self.prefix_tabs(ids.splitlines())

    def get_words(self, tab_ids, match_regex, join_with):
        words = set()
        match_regex = encode_query(match_regex)
        join_with = encode_query(join_with)

        for tab_id in tab_ids:
            prefix, _window_id, tab_id = tab_id.split('.')
            if prefix + '.' != self._prefix:
                continue

            logger.info(
                'SingleMediatorAPI: get_words: %s, match_regex: %s, join_with: %s',
                tab_id, match_regex, join_with)
            words |= set(self._get(
                '/get_words/%s?match_regex=%s&join_with=%s' % (tab_id, match_regex, join_with)
            ).splitlines())

        if not tab_ids:
            words = set(self._get(
                '/get_words?match_regex=%s&join_with=%s' % (match_regex, join_with)
            ).splitlines())

        return sorted(list(words))

    def get_text_or_html(self, command, args, delimiter_regex, replace_with):
        num_tabs = MAX_NUMBER_OF_TABS
        if len(args) > 0:
            num_tabs = int(args[0])

        result = self._get(
            '/%s?delimiter_regex=%s&replace_with=%s' % (
                command,
                encode_query(delimiter_regex),
                encode_query(replace_with),
            ),
        )
        lines = []
        for line in result.splitlines()[:num_tabs]:
            lines.append(line)
        return self.prefix_tabs(lines)

    def get_text(self, args, delimiter_regex, replace_with):
        return self.get_text_or_html('get_text', args, delimiter_regex, replace_with)

    def get_html(self, args, delimiter_regex, replace_with):
        return self.get_text_or_html('get_html', args, delimiter_regex, replace_with)

    def shutdown(self):
        return self._get('/shutdown')

    def _get(self, path, data=None):
        return self._client.get(path, data)

    def _post(self, path, files=None):
        return self._client.post(path, files)


def api_must_ready(port: int, browser: str,
                   prefix='a',
                   client_timeout: float = 0.1,
                   startup_timeout: float = 1.0) -> SingleMediatorAPI:
    client = HttpClient(DEFAULT_HTTP_IFACE, port, timeout=client_timeout)
    api = SingleMediatorAPI(prefix=prefix, port=port, startup_timeout=startup_timeout, client=client)
    assert api.browser == browser
    return api


def int_tab_id(tab_id: str) -> int:
    """Convert from str(b.20.123) to int(123)"""
    return int(tab_id.split('.')[-1])

class MultipleMediatorsAPI(object):
    """
    This API is designed to work with multiple mediators.
    """

    def __init__(self, apis):
        self._apis = apis

    @property
    def ready_apis(self):
        return [api for api in self._apis if api.ready]

    def close_tabs(self, args):
        for api in self._apis:
            api.close_tabs(args)

    def activate_tab(self, args: List[str], focused: bool):
        if len(args) == 0:
            print('Usage: brotab_client.py activate_tab [--focused] <#tab>')
            return 2

        for api in self._apis:
            api.activate_tab(args, focused)

    def get_active_tabs(self, args):
        return [api.get_active_tabs(args) for api in self._apis]

    def query_tabs(self, args, print_error=False):
        functions = [partial(api.query_tabs_safe, args, print_error)
                     for api in self.ready_apis]
        if not functions:
            return []
        tabs = sum(call_parallel(functions), [])
        return tabs

    def list_tabs(self, args, print_error=False):
        functions = [partial(api.list_tabs_safe, args, print_error)
                     for api in self.ready_apis]
        if not functions:
            return []
        tabs = sum(call_parallel(functions), [])
        return tabs

    def _move_tabs_if_changed(self, api, tabs_before, tabs_after):
        delete_commands, move_commands, update_commands = infer_all_commands(
            parse_tab_lines(tabs_before),
            parse_tab_lines(tabs_after))

        if delete_commands:
            print('DELETE', delete_commands)
            api.close_tabs(delete_commands)

        if move_commands:
            print('MOVE', move_commands)
            api.move_tabs(move_commands)

        if update_commands:
            print('UPDATE', update_commands)
            api.update_tabs(update_commands)

    def update_tabs(self, all_updates):
        results = []
        for api in self._apis:
            updates = [deepcopy(u) for u in all_updates if api.prefix_match(u['tab_id'])]
            for u in updates:
                u['tab_id'] = int_tab_id(u['tab_id'])
            results.extend(api.update_tabs(updates))
        return results

    def move_tabs(self, args):
        """
        This command allows closing tabs and move them around.

        It lists current tabs, opens an editor, and when editor is done, it
        detects which tabs where deleted and which where moved. It closes/
        removes tabs, and moves the rest accordingly.

        Alg:
        1. find maximum sequence of ordered tabs in the output
        2. take another tab from the input,
            - find a position in the output where to put that new tab,
              using binary search
            - insert that tab
        3. continue until no input tabs out of order are left
        """
        tabs_before = self.list_tabs(args, print_error=True)
        tabs_after = edit_tabs_in_editor(tabs_before)
        if tabs_after is None:
            return

        for api in self._apis:
            self._move_tabs_if_changed(
                api,
                api.filter_tabs(tabs_before),
                api.filter_tabs(tabs_after))

    def _get_api_by_prefix(self, prefix):
        for api in self._apis:
            if api._prefix == prefix:
                return api
        raise ValueError('No such client with prefix "%s"' % prefix)

    def open_urls(self, urls, prefix, window_id=None):
        assert len(self._apis) > 0, \
            'There should be at least one client connected: %s' % self._apis
        # client = self._apis[0]
        client = self._get_api_by_prefix(prefix)
        return client.open_urls(urls, window_id)

    def get_words(self, tab_ids, match_regex, join_with):
        words = set()
        import time
        for api in self.ready_apis:
            start = time.time()
            words |= set(api.get_words(tab_ids, match_regex, join_with))
            delta = time.time() - start
            # print('DELTA', delta, file=sys.stderr)
        return sorted(list(words))

    def _get_text_or_html(self, api, getter, args, delimiter_regex, replace_with):
        result = []
        try:
            import time
            start = time.time()
            result = getter(args, delimiter_regex, replace_with)
            delta = time.time() - start
            logger.info('get text/html (single client) took %s', delta)
        except ValueError as e:
            print("Cannot decode JSON: %s: %s" % (api, e), file=sys.stderr)
            logger.error("Cannot decode JSON: %s: %s" % (api, e))
        except URLError as e:
            print("Cannot access API %s: %s" % (api, e), file=sys.stderr)
            logger.error("Cannot access API %s: %s" % (api, e))
        except Exception as e:
            logger.error("Unknown exception: %s %s" % (api, e))
        return result

    def get_text(self, args, delimiter_regex, replace_with):
        tabs = []
        for api in self.ready_apis:
            tabs.extend(self._get_text_or_html(api, api.get_text, args,
                                               delimiter_regex, replace_with))
        return tabs

    def get_html(self, args, delimiter_regex, replace_with):
        tabs = []
        for api in self.ready_apis:
            tabs.extend(self._get_text_or_html(api, api.get_html, args,
                                               delimiter_regex, replace_with))
        return tabs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command
import pip

# Save ~200ms on script startup time
# See https://github.com/ninjaaron/fast-entry_points
import fastentrypoints

# Package meta-data.
NAME = 'brotab'
DESCRIPTION = "Control your browser's tabs from the command line"
URL = 'https://github.com/balta2ar/brotab'
EMAIL = 'baltazar.bz@gmail.com'
AUTHOR = 'Yuri Bochkarev'


# What packages are required for this module to be executed?
REQUIRED = [
    # 'requests', 'maya', 'records',
]
# requirements = list(pip.req.parse_requirements(
#     'requirements.txt', session=pip.download.PipSession()))
# REQUIRED = [requirement.name for requirement in requirements]
REQUIRED = [line.strip() for line in open('requirements/base.txt').readlines()]

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
# with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
#     long_description = '\n' + f.read()

try:
    #import pypandoc
    #long_description = pypandoc.convert_file("README.md", "rst")
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except ImportError:
    print('WARNING: description is empty, consider installing pypandoc: pip install --user pypandoc')
    long_description = ''

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


packages = find_packages(
    # where='brotab',
    # exclude=('brotab.tests', 'firefox_extension', 'firefox_mediator')
    #exclude=('tests', 'firefox_extension', 'firefox_mediator')
    include=(
        'brotab',
        'brotab.tests',
        'brotab.search',
        'brotab.mediator',
        'brotab.albert',
    ),
    exclude=('firefox_extension', 'firefox_mediator')
)
print('>>', packages)


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=packages,
    # packages=find_packages(
    #     where='brotab',
    #     # exclude=('brotab.tests', 'firefox_extension', 'firefox_mediator')
    #     exclude=('tests', 'firefox_extension', 'firefox_mediator')
    # ),
    data_files=[
        ('config', ['brotab/mediator/chromium_mediator.json',
                    'brotab/mediator/firefox_mediator.json']),
    ],
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],

    entry_points={
        'console_scripts': [
            'brotab=brotab.main:main',
            'bt=brotab.main:main',
            'bt_mediator=brotab.mediator.brotab_mediator:main',
        ],
    },
    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



#!/usr/bin/env python3

import logging
import logging.handlers
import os
import re
import socket

from brotab.inout import get_mediator_ports
from brotab.inout import is_port_accepting_connections
from brotab.mediator import sig
from brotab.mediator.const import DEFAULT_HTTP_IFACE
from brotab.mediator.const import DEFAULT_SHUTDOWN_POLL_INTERVAL
from brotab.mediator.http_server import MediatorHttpServer
from brotab.mediator.log import disable_click_echo
from brotab.mediator.log import mediator_logger
from brotab.mediator.remote_api import default_remote_api
from brotab.mediator.transport import default_transport


# TODO:
# 1. Run HTTP server and accept the following commands:
#    - /list_tabs
#    - /close_tabs
#    - /move_tabs ???
#    - /open_tab
#    - /open_urls
#    - /new_tab (google search)
#    - /get_tab_text
#    - /get_active_tabs_text
#
# TODO: fix bug when the number of tabs > 1100
# TODO: read stdin continuously in a separate thread,
#       detect if it's closed, shutdown the server, and exit.
#       make sure this threaded reader and server reader are mutually exclusive.
# TODO: all commands should be synchronous and should only terminate after
#       the action has been actually executed in the browser.
# TODO: logs from main and mediator should go into different files
# TODO: bt html may cause "Uncaught (in promise) Error: Message length exceeded maximum allowed length."


def monkeypatch_socket_bind_allow_port_reuse():
    """Allow port reuse by default"""
    socket.socket._bind = socket.socket.bind

    def my_socket_bind(self, *args, **kwargs):
        mediator_logger.info('Custom bind called: %s, %s', args, kwargs)
        self.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return socket.socket._bind(self, *args, **kwargs)

    socket.socket.bind = my_socket_bind


def blacklist_loggers():
    blacklist = r'.*pyppeteer.*|.*urllib.*|.*socks.*|.*requests.*|.*dotenv.*'
    for name in logging.root.manager.loggerDict:
        match = re.match(blacklist, name) is not None
        # print(name, match)
        if match:
            logger = logging.getLogger(name)
            logger.setLevel(logging.ERROR)
            logger.propagate = False


def mediator_main():
    monkeypatch_socket_bind_allow_port_reuse()
    disable_click_echo()

    port_range = list(get_mediator_ports())
    transport = default_transport()
    # transport = transport_with_timeout(sys.stdin.buffer, sys.stdout.buffer, DEFAULT_TRANSPORT_TIMEOUT)
    # transport = transport_with_timeout(sys.stdin.buffer, sys.stdout.buffer, 1.0)
    remote_api = default_remote_api(transport)
    host = DEFAULT_HTTP_IFACE
    poll_interval = DEFAULT_SHUTDOWN_POLL_INTERVAL

    for port in port_range:
        mediator_logger.info('Starting mediator on %s:%s...', host, port)
        if is_port_accepting_connections(port):
            continue
        try:
            server = MediatorHttpServer(host, port, remote_api, poll_interval)
            thread = server.run.in_thread()
            sig.setup(lambda: server.run.shutdown(join=False))
            # server.run.parent_watcher(thread.is_alive, interval=1.0)
            thread.join()
            mediator_logger.info('Exiting mediator pid=%s on %s:%s...', os.getpid(), host, port)
            break
        except OSError as e:
            # TODO: fixme: we won't get this if we run in a process
            mediator_logger.info('Cannot bind on port %s: %s', port, e)
        except BrokenPipeError as e:
            # TODO: probably also won't work with processes, also a race
            mediator_logger.exception('Pipe has been closed (%s)', e)
            server.run.shutdown(join=True)
            break

    else:
        mediator_logger.error('No TCP ports available for bind in range %s', port_range)


def main():
    mediator_main()


if __name__ == '__main__':
    main()



import os
import signal
from contextlib import contextmanager
from multiprocessing import Barrier
from multiprocessing import Process
from unittest import TestCase

from brotab.api import api_must_ready
from brotab.inout import get_available_tcp_port
from brotab.mediator import sig
from brotab.mediator.const import DEFAULT_HTTP_IFACE
from brotab.mediator.http_server import MediatorHttpServer
from brotab.mediator.log import mediator_logger
from brotab.mediator.remote_api import default_remote_api
# from brotab.mediator.transport import transport_with_timeout
from brotab.wait import Waiter


# class MockedPiperMediator:
#     def __init__(self, prefix='a', port=None, remote_api=None):
#         mediator_logger.info('starting mediator pid=%s', os.getpid())
#         self.prefix = prefix
#         self.port = get_available_tcp_port() if port is None else port
#         input_r, input_w = os.pipe()
#         output_r, output_w = os.pipe()
#         self.transport_browser = transport_with_timeout(output_r, input_w, 0.050)
#         self.transport_mediator = transport_with_timeout(input_r, output_w, 0.050)
#         self.remote_api = default_remote_api(self.transport_mediator) if remote_api is None else remote_api
#         self.server = MediatorHttpServer(DEFAULT_HTTP_IFACE, self.port, self.remote_api, poll_interval=0.050)
#         self.thread = None
#         self.api = None
#
#     def start(self):
#         self.thread = self.server.run.in_thread()
#         self.transport_browser.send('mocked')
#
#     def wait_api_ready(self):
#         self.api = api_must_ready(port=self.port, browser='mocked', prefix=self.prefix,
#                                   client_timeout=0.1, startup_timeout=1)
#
#     def join(self):
#         # sig.setup(lambda: server.run.shutdown(join=False))
#         self.server.run.parent_watcher(self.thread.is_alive, interval=0.050)  # this is crucial
#         self.thread.join()
#
#     def __enter__(self):
#         return self
#
#     def __exit__(self, type_, value, tb):
#         self.join()
#
#
# class TestMediatorThreadTerminates(TestCase):
#     def setUp(self):
#         self.mediator = MockedPiperMediator()
#         self.mediator.start()
#         self.mediator.wait_api_ready()
#
#     def tearDown(self):
#         pass
#
#     def test_if_cannot_read(self):
#         self.mediator.api.list_tabs([])
#         self.mediator.join()  # this should complete without manual shutdown
#
#     def test_if_cannot_write(self):
#         self.mediator.transport_browser.send(['1.1\ttitle\turl'])  # make reads work
#         self.mediator.transport_browser.close()
#         self.mediator.api.list_tabs([])
#         self.mediator.join()  # this should complete without manual shutdown
#
#
# class TestMediatorProcessTerminates(TestCase):
#     def test_when_parent_died(self):
#         port = get_available_tcp_port()
#         mediator_logger.info('starting test pid=%s', os.getpid())
#         kill_parent = Barrier(2)
#
#         def run_threaded_mediator():
#             mediator = MockedPiperMediator(port=port)
#             mediator.start()
#             mediator.join()
#
#         def run_doomed_parent_browser():
#             mediator_logger.info('doomed_parent_browser pid=%s', os.getpid())
#             mediator_process = Process(target=run_threaded_mediator)
#             mediator_process.start()
#             mediator_process.join()
#
#         def on_sig_child(signum, frame):
#             pid, status = os.wait()
#             mediator_logger.info('reaped child signum=%s pid=%s status=%s', signum, pid, status)
#
#         def run_supervisor():
#             signal.signal(signal.SIGCHLD, on_sig_child)
#             doomed_parent_browser = Process(target=run_doomed_parent_browser)
#             doomed_parent_browser.start()
#             kill_parent.wait()
#             doomed_parent_browser.terminate()
#             doomed_parent_browser.join()
#
#         signal.signal(signal.SIGCHLD, on_sig_child)
#         supervisor = Process(target=run_supervisor)
#         supervisor.start()
#         api = api_must_ready(port, 'mocked')
#
#         # kill parent and expect mediator to terminate as well
#         kill_parent.wait()
#         self.assertTrue(Waiter(api.pid_not_ready).wait(timeout=1.0))
#         supervisor.join()
#
#     @contextmanager
#     def run_as_child_process(self):
#         port = get_available_tcp_port()
#         mediator_logger.info('starting test pid=%s', os.getpid())
#
#         def run_threaded_mediator():
#             mediator = MockedPiperMediator(port=port)
#             mediator.start()
#             sig.setup(lambda: mediator.server.run.shutdown(join=False))
#             mediator.join()
#
#         mediator_process = Process(target=run_threaded_mediator)
#         mediator_process.start()
#         api = api_must_ready(port, 'mocked')
#
#         yield mediator_process
#
#         self.assertTrue(Waiter(api.pid_not_ready).wait(timeout=1.0))
#         mediator_process.join()
#
#     def test_when_sigint_received(self):
#         with self.run_as_child_process() as mediator_process:
#             os.kill(mediator_process.pid, signal.SIGINT)
#
#     def test_when_sigterm_received(self):
#         with self.run_as_child_process() as mediator_process:
#             os.kill(mediator_process.pid, signal.SIGTERM)



"""
This module contains helpers that query the indexed database of text from a
browser.
"""

import argparse
import sqlite3
import logging
from collections import namedtuple

logger = logging.getLogger('brotab')

QueryResult = namedtuple('QueryResult', 'tab_id title snippet')


def query(sqlite_filename, user_query,
          text_column_index=3,
          max_tokens=30,
          max_results=10,
          marker_start='<b>',
          marker_end='</b>',
          marker_cut='...'):
    logger.info('Executing sqlite db %s query "%s"',
                sqlite_filename, user_query)
    conn = sqlite3.connect(sqlite_filename)
    cursor = conn.cursor()
    query = """
    select
        rank,
        tab_id,
        title,
        snippet(tabs,
                {text_column_index},
                '{marker_start}',
                '{marker_end}',
                '{marker_cut}',
                {max_tokens}) body
    from tabs where tabs match ? order by rank limit {max_results};
""".format_map(locals())
    results = []
    try:
        for (_rank, tab_id, title, snippet) in cursor.execute(query, (user_query,)):
            results.append(QueryResult(tab_id, title, snippet))
    except sqlite3.OperationalError as e:
        logger.exception('Error: %s', e)

    conn.close()
    return results


def main():
    parser = argparse.ArgumentParser(description='Query text DB')
    parser.add_argument('sqlite', help='sqlite DB filename')
    parser.add_argument('query', help='sqlite query')
    args = parser.parse_args()

    for result in query(args.sqlite, args.query):
        print('\t'.join([result.tab_id, result.title, result.snippet]))


if __name__ == '__main__':
    main()
